// Description: a node that generates events with the value of the microphone amplitude
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> the volume of a sound is the region in which the sound will play at 100% amplitude(loudness)
//<li> falloffDistance is a region outside the volume in which the amplitude is diminished as the user moves further away from the source(volume)
//<li> the sound server, snerd, must be run on the a local or remote machine
//<li> sounds under a switch will turn off but must be restarted when the switch is toggled back on
//</ul>
//
// Category: Sound
// Author: Alex Hill
// Revision: 08/01/03
//
#include <pfcave.h>
#include <bergenAmplitude.h>
#include <ygWorld.h>
#include <ygNetMessages.h>
#include "amplitude.h"

using namespace std;

extern "C" ygNode* APIENTRY construct_amplitude(const char* name,bool master) { return new amplitude(name,master); }

struct _amplitudePrivateData
    {
    bergenAmplitude* sample;
    bool isPlaying;
    bool netPlaying;
    bool isPaused;
    float inputAmplitude;
    float threshold;
    float filter;
    float curAmplitude;
    float distance;
    pfVec3 currPosition;
    float updateInterval, lastUpdate;
    };


amplitude::amplitude(const char* name,bool master) : ygSoundServer(name,master)
    {
    setClassName("amplitude");
    p_ = new struct _amplitudePrivateData;
    p_->sample = NULL;
    p_->isPlaying = false;
    p_->isPaused = false;
    p_->netPlaying = false;
    p_->inputAmplitude = 0;
    p_->curAmplitude = 0;
    p_->threshold = 0.2;
    p_->filter = 0.8;
    p_->updateInterval = 0.1;
    if (getenv("YG_SOUND_INTERVAL"))
        p_->updateInterval = atof(getenv("YG_SOUND_INTERVAL"));
    p_->lastUpdate = 0;
    addNetMessage("play", &p_->netPlaying, YG_NET_BOOL_NEGATE);
    addNetMessage("stop", &p_->netPlaying, YG_NET_BOOL);
    dontSaveNetMessage("stop");
    addNetMessage("threshold",&p_->threshold,YG_NET_FLOAT);
    addNetMessage("filter",&p_->filter,YG_NET_FLOAT);
    addNetMessage("updateInterval",&p_->updateInterval,YG_NET_FLOAT);
    addEvent("changed");
    }

amplitude::~amplitude(void)
    {
    if (p_->sample)
        delete p_->sample;
    delete p_;
    }


void amplitude::reset(void)
    {
    //reset playing flag to false
    p_->isPlaying = false;
    p_->isPaused = false;
    //set threshold to 0.2
    p_->threshold = 0.2;
    //set filter to 0.8
    p_->filter = 0.8;
    //set current amplitude to 0.0
    p_->curAmplitude = 0;
    p_->inputAmplitude = 0;
    //set update interval to 0.05
    p_->updateInterval = 0.05;
    //if a sample exists then delete it
    if (p_->sample)
        {
        stop();
        delete p_->sample;
        p_->sample = NULL;
        }
    ygSoundServer::reset();
    }


void amplitude::message(const ygMessage& msg)
    {
    //set the minimum amplitude threshold
    if (msg == "threshold")
        {
        return;
        }
    //set the first order filter constant (percent contribution)
    else if (msg == "filter")
        {
        return;
        }
    //start monitoring the input signal
    else if (msg == "play")
        {
        if (p_->netPlaying)
            play();
        else
            stop();
        }
    //stop the playing sound
    else if (msg == "stop")
        {
        if (p_->netPlaying)
            play();
        else
            stop();
        }
    //set the update interval for sound server
    else if (msg == "updateInterval")
        {
        return;
        }
    else
        ygSoundServer::message(msg);
    }


void amplitude::setDistance(float val)
    {
    if (p_->sample)
        p_->sample->setDistance(val);
    p_->distance = val;
    }


void amplitude::setAmplitude(float val)
    {
    if (p_->sample)
        p_->sample->setAmplitude(val);
    ygSoundServer::setAmplitude(val);
    }


void amplitude::setFalloffDistance(float val)
    {
    if (p_->sample)
        p_->sample->setFalloffDistance(val);
    ygSoundServer::setFalloffDistance(val);
    }


void amplitude::setAttenuation(int val)
    {
    if (p_->sample)
        p_->sample->setAttenuationType((AttenuationType)val);
    ygSoundServer::setAttenuation(val);
    }


bool amplitude::isPlaying(void)
    {
    return p_->isPlaying;
    }


bool amplitude::isPaused(void)
    {
    return p_->isPaused;
    }


void amplitude::play(void)
    {
    if (p_->isPlaying)
        return;
    //if no sample then create a new bergenAmplitude
    if (!p_->sample)
        p_->sample = new bergenAmplitude(server());
    p_->sample->setAmplitude(ygSoundServer::amplitude());
    p_->sample->setFalloffDistance(falloffDistance());
    p_->sample->setAttenuationType((AttenuationType)attenuation());
    setDistance(computeDistance());
    //play the sample
    p_->sample->play();
    //generate a heartbeat for the server
    heartbeat();
    p_->isPlaying = true;
    eventOccurred("start");
    }


void amplitude::pause(void)
    {
    if (p_->isPlaying)
        {
        p_->isPaused = true;
        p_->isPlaying = false;
        p_->sample->pause();
        }
    }


void amplitude::stop(bool sendStopCommand)
    {
    //if a sample is playing
    if (p_->isPlaying || p_->isPaused)
        {
        //send a stop command to the server
        if ((sendStopCommand) && (p_->sample))
            p_->sample->stop();
        p_->isPlaying = false;
        p_->isPaused = false;
        p_->netPlaying = false;
        eventOccurred("stop");
        }
    }


void amplitude::app(void)
    {
    //if a sample exists
    if (p_->sample)
        {
        //generate a heartbeat for the server
        heartbeat();
        //if the sample is playing
        if (p_->isPlaying)
            {
            //if update interval has passed then update sound server
            if ((ygWorld::frameTime() - p_->lastUpdate) > p_->updateInterval)
                {
                float currAmplitude;
                currAmplitude = p_->sample->amplitude();
                if (currAmplitude < p_->threshold)
                    currAmplitude = 0.0;
                currAmplitude = currAmplitude*p_->filter + p_->inputAmplitude*(1-p_->filter);
                if (currAmplitude != p_->inputAmplitude)    
                    {
                    p_->inputAmplitude = currAmplitude;
                    ygString args("value=");
                    char val[32];
                    sprintf(val,"%f",p_->inputAmplitude);
                    args += val;
                    //generate changed event with $value
                    eventOccurred("changed",args);
                    }
                float newDistance = computeDistance();
                if (newDistance != p_->distance)
                    setDistance(newDistance);
                //determine position of this sound
                pfVec3 newPos;
                newPos = origin();
                pfVec3 cavePos;
                //get position relative to CAVE
                CAVENavConvertWorldToCAVE(newPos.vec,cavePos.vec);
                if (cavePos != p_->currPosition)
                    {
                    if (p_->sample)
                        p_->sample->setPosition(cavePos.vec[0],cavePos.vec[2],-cavePos.vec[1]);
                    p_->currPosition = cavePos;
                    }
                p_->lastUpdate = ygWorld::frameTime();
                }
            }
        }
    ygSoundServer::app();
    }
