// Description: generates events describing node class types
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//</ul>
//
// Category: System
// Author: Alex Hill
// Revision: 09/15/05 Alex Hill
//
#include <sys/types.h>
#include <sys/stat.h>
#ifndef WIN32
#include <dirent.h>
#endif
#include <ygNodeDB.h>
#include <ygNetMessages.h>
#include <ygDSOLoader.h>
#include "classListing.h"

using namespace std;

extern "C" ygNode* APIENTRY construct_classListing(const char* name,bool master) { return new classListing(name,master); }

classListing::classListing(const char* name,bool master) : ygNode(name,master)
    {
    setClassName("classListing");
    index = 0;
    maximum = 100;
    created = true;
    node = NULL;
    users = true;
    addNetMessage("node",&nodeName,YG_NET_NODE);
    addNetMessage("node",&index,YG_NET_INT);
    addNetMessage("created",&created,YG_NET_BOOL);
    addNetMessage("maximum",&maximum,YG_NET_INT);
    addNetMessage("users",&users,YG_NET_BOOL);
    addEvent("node");
    addEvent("changed");
    vector<ygString> excluded;
    excluded.push_back("Node");
    excluded.push_back("ygGroup");
    excluded.push_back("Group");
    excluded.push_back("Transform");
    excluded.push_back("ygSimpleTransform");
    excluded.push_back("SimpleTransform");
    excluded.push_back("Light");
    excluded.push_back("Object");
    excluded.push_back("staticObject");
    excluded.push_back("Texture");
    excluded.push_back("Switch");
    excluded.push_back("Selector");
    excluded.push_back("User");
    excluded.push_back("Head");
    excluded.push_back("Wand");
    excluded.push_back("ygSoundServer");
    excluded.push_back("Sound");
    excluded.push_back("Space");
    excluded.push_back("scene");
    excluded.push_back("Environment");
    excluded.push_back("geometry");
    excluded.push_back("World");
    excluded.push_back("CAVENavigator");
    excluded.push_back("ygCAVEHead");
    excluded.push_back("CAVEHead");
    excluded.push_back("CAVETracker");
    excluded.push_back("CAVEWand");
    const map<ygString,class ygDSOInfo*>& infomap = ygDSOLoader::getDSOInfoMap();
    map<ygString,class ygDSOInfo*>::const_iterator i;
    for (i=infomap.begin(); i != infomap.end(); ++i)
        {
        ygString currType((*i).first);
        bool include=true;
        vector<ygString>::iterator i;
        for (i=excluded.begin();i!=excluded.end();i++)
            {
            if (currType == (*i))
                include = false;
            }
        if (include)
            {
            vector<ygString>::iterator i;
            //order allClasses in alphabetical order
            for (i=allClasses.begin();i!=allClasses.end();i++)
                {
                if (currType < (*i))
                    break;
                }
            //insert the new directory into the list
            allClasses.insert(i,currType);
            }
        }    
    }
    
void classListing::reset(void)
    {
    node = NULL;
    lastIndex = -1;
    //clear the start node name
    nodeName.clear();
    //set the created flag to true
    created = true;
    //set the maximum printout to 100
    maximum = 100;
    //set the index to 0
    index = 0;
    //set the users flag to true
    users = true;
    currClasses.clear();
    ygNode::reset();
    }

void classListing::message(const ygMessage& msg)
    {
    //generate events with class names from the given node and index
    if (msg == "node")
        {
        checkNode();
        }
    //generate events with class names using the given index
    else if (msg == "index")
        {
        if (msg.numArgs() > 0)
            {
            index = msg.intArg(0);
            lastIndex = -1;
            checkNode();
            }
        else
            msg.error(name(),"(wrong number of arguments)");
        }
    //set the flag to display only existing nodes
    else if (msg == "created")
        {
        return;
        }
    //set the total number of classes to report
    else if (msg == "maximum")
        {
        return;
        }
    //set flag to include ygUser class in the list
    else if (msg == "users")
        {
        return;
        }
    else
        ygNode::message(msg);
    }


void classListing::findClasses(ygNode* currNode)
    {
    ygString currClass;
    currClass = currNode->className();
    bool duplicate=false;
    vector<ygString>::iterator i;
    //order currClasses in alphabetical order
    for (i=currClasses.begin();i!=currClasses.end();i++)
        {
        if (currClass == (*i))
            duplicate = true;
        if (currClass < (*i))
            break;
        }
    //insert the new directory into the list
    if (!duplicate)
        currClasses.insert(i,currClass);    
    //call findNode recursively on all children
    for (int i=0; i < currNode->numChildren(); i++)
        findClasses(currNode->child(i));
    }


void classListing::findUsers(void)
    {
    ygString currClass;
    currClass = "ygUser";
    bool duplicate=false;
    vector<ygString>::iterator i;
    //order currClasses in alphabetical order
    for (i=currClasses.begin();i!=currClasses.end();i++)
        {
        if (currClass == (*i))
            duplicate = true;
        if (currClass < (*i))
            break;
        }
    //insert the new directory into the list
    if (!duplicate)
        currClasses.insert(i,currClass);    
    }


void classListing::app(void)
    {
    if (node && node->isChanged(YG_CHANGE_TREE))
        {
        //populate current class list
        currClasses.clear();
        findClasses(node);
        if (users)
            findUsers();
        lastIndex = -1;
        checkNode();
        }
    ygNode::app();
    }


void classListing::checkNode(void)
    {
    if (nodeName != lastNode)
        {
        if (nodeName.length() > 0)
            node = ygNodeDB::find(nodeName,false);
        else
            node = NULL;
        if (node)
            {
            currClasses.clear();
            findClasses(node);
            if (users)
                findUsers();
            lastIndex = -1;
            }
        lastNode = nodeName;
        }
    if (index != lastIndex)
        {
        int classCount;
        if (created)
            classCount = currClasses.size();
        else
            classCount = allClasses.size();
        int classMax = index + maximum;
        if (classMax > classCount)
            classMax = classCount;
        ygString args("total=");
        char val[32];
        sprintf(val,"%d",classCount);
        args += val;
        eventOccurred("changed",args);
        int num=0;
        for (int i=index;i<classMax;i++)
            {
            ygString args("name=");
            if (created)
                args += currClasses[i];
            else
                args += allClasses[i];
            args += " number=";
            sprintf(val,"%d",num);
            args += val;
            eventOccurred("node",args);
            num++;
            }
        }
    lastIndex = index;
    }
