// Description: replaces the texture on objects below it with that of a movie file
//
//<b>notes:</b>
//<ul>
//<li> all parent node messages are inherited
//<li> see reset method for default settings
//<li> the Quicktime (*.mov) photoJPEG format is only supported on Linux and requires the openquicktime library
//<li> you must add the following to your RUN file in order to use this node on Linux:
//<ul>
//<li> setenv LD_LIBRARY_PATH /usr/local/lib
//</ul>
//</ul>
//
// Category: Attributes
// Author: Dave Pape
//           11/10/01
// Revision: 01/10/03 Alex Hill - added Linux support thanks to Ben Chang
//
#include <malloc.h>
#include <Performer/pf/pfGroup.h>
#include <Performer/pf/pfGeode.h>
#include <Performer/pr/pfGeoSet.h>
#include <Performer/pr/pfTexture.h>
#include <GL/glext.h>
#include "ygUtil.h"
#include "ygWorld.h"
#include "ygNetMessages.h"
#include "movieTexture.h"
#include "movieObject.h"


using namespace std;

extern "C" ygNode* APIENTRY construct_movieTexture(const char* name,bool master) { return new movieTexture(name,master); }


struct _movieTexturePrivateData
    {
    pfTexture* texture;
    bool active;
    float startTime;
    int width;
    int height;
    float rate;
    unsigned char* buffer; 
    movieObject movie;
    ygString netFile;
    bool debugFrame;
    };

movieTexture::movieTexture(const char* name,bool master) : ygNode(name,master)
    {
    setClassName("movieTexture",true);
    p_ = new struct _movieTexturePrivateData;
    p_->active = false;
    p_->texture = new pfTexture;
    p_->buffer = (unsigned char*)pfMalloc(256*256*3,pfGetSharedArena());
    memset(p_->buffer,0,256*256*3);
    p_->texture->setFormat(PFTEX_IMAGE_FORMAT, PFTEX_RGB);
    p_->texture->setImage((uint*)p_->buffer,3,256,256,1);
    addNetMessage("file",&p_->netFile,YG_NET_FILE);
    addNetMessage("movie",&p_->netFile,YG_NET_FILE);
    dontSaveNetMessage("movie");
    addNetMessage("play",&p_->active,YG_NET_BOOL_NEGATE);
    debugFlag("frame",&p_->debugFrame);
    addEvent("end");
    }

movieTexture::~movieTexture(void)
    {
    delete p_;
    }


void movieTexture::reset(void)
    {
    ygNode::reset();
    //set active to false
    resetNetMessage("play");
    //reset movie file
    resetNetMessage("file");
    //turn off frame debug
    p_->debugFrame = false;
    }


void movieTexture::message(const ygMessage& msg)
    {
    //set the name of the movie file
    if (msg == "file" || msg == "movie")
        {
        if (p_->netFile.length() > 0)
            setMovie(p_->netFile);
        }
    //start the movie from the beginning
    else if (msg == "play")
        {
        if (p_->active == true)
            p_->startTime = ygWorld::frameTime();
        }
    //stop the movie
    else if (msg == "stop")
        {
        p_->active = false;
        netMessageChanged("play");
        }
    else
        ygNode::message(msg);
    }


void movieTexture::setMovie(const ygString& file)
    {
    char *searchPath = ".";
    ygString fullPath;
    if (getenv("YG_PATH"))
        searchPath = getenv("YG_PATH");
    if (!ygFindFile(file,searchPath,fullPath))
        {
        fprintf(stderr,"ERROR: movieTexture::show: could not find file '%s'\n",file.c_str());
        return;
        }
    if (p_->movie.isOpen())
        p_->movie.close();
    p_->movie.open(fullPath.c_str());
    if (!p_->movie.isOpen())
        {
        fprintf(stderr,"ERROR (movieTexture::show): failed to open '%s'\n",fullPath.c_str());
        return;
        }
    p_->width = p_->movie.getWidth();
    p_->height = p_->movie.getHeight();
    p_->rate = p_->movie.getRate();
    pfDelete(p_->buffer);
    p_->buffer = (unsigned char*)pfMalloc(p_->width*p_->height*4*sizeof(unsigned char),pfGetSharedArena());
    p_->netFile = file;
    }

void movieTexture::app(void)
    {
    if (isChanged(YG_CHANGE_GEOMETRY | YG_CHANGE_TEXTURE))
        replaceTextures(pfnode());
    if (p_->active)
        {
        int frame = (int)((ygWorld::frameTime() - p_->startTime)*p_->rate);        
        if (p_->debugFrame)
            cout << "Reading frame " << frame << " of " << p_->netFile << endl;
        if (p_->movie.getFrame(frame,p_->buffer))
            {
            p_->texture->setFormat(PFTEX_IMAGE_FORMAT, GL_ABGR_EXT);
            p_->texture->setImage((uint*)p_->buffer,4,p_->width,p_->height,1);
            }
        else
            {
            p_->active = false;
            netMessageChanged("play");
            //the movie has ended
            eventOccurred("end");
            }
        }
    ygNode::app();
    }

void movieTexture::replaceTextures(pfNode * node)
    {
    int i;
    if (node->isOfType(pfGeode::getClassType()))
        {
        pfGeode *geode = (pfGeode *)node;
        for (i=0; i < geode->getNumGSets(); i++)
            {
            pfGeoSet *gset = geode->getGSet(i);
            pfGeoState* gstate = gset->getGState();
            pfTexture* texture;
            texture = (pfTexture*)gstate->getAttr(PFSTATE_TEXTURE);
            if (texture)
                gstate->setAttr(PFSTATE_TEXTURE, p_->texture);
            }
        }
    else if (node->isOfType(pfGroup::getClassType()))
        {
        pfGroup *group = (pfGroup *)node;
        for (i=0; i < group->getNumChildren(); i++)
            replaceTextures(group->getChild(i));
        }
    }

