// Description: a node that creates a particle system
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//</ul>
//
// Category: Geometry
// Author: Sangyoon Lee
//           11/01/03
// Revision: 03/07/06 Alex Hill - converted to YG 4.X and added networking
//
#include <Performer/pf/pfGroup.h>
#include <Performer/pf/pfGeode.h>
#include <Performer/pr/pfGeoSet.h>
#include <Performer/pr/pfTexture.h>
#include <Performer/pr/pfMaterial.h>
#include <Performer/pf/pfBillboard.h>
#include "particleSystem.h"
#include <ygNetMessages.h>
#include "ygUtil.h"
#include <ygUser.h>
#include <ygWorld.h>
#include <ygPFTextureCache.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// define max length of string
#define MAX_LENGTH 256

#define SERVERMODE 0
using namespace std;

#ifdef WIN32
#include <stdlib.h>
#include <time.h>

#define drand48()   ((double) ((double) rand() / (double) RAND_MAX))
#endif

#define GRAVITY                pfVec3(0.0f, 0.0f, -9.8f)
#define PT_PI                ((float)  3.141592654f)
#define ROOT2INV        0.7071067811865475244008443621052f
#define MAXPARTICLENUM 200 
#define I_MAX_STRING_LENGTH 250
#define RandomIn(a, b)    (a + (float)drand48() * (b - a))
#define Degree2Radian(a) ((float) a * PT_PI / 180.0f)
#define CM2Feet(a) ((float) a / 30.0f)

extern "C" ygNode* APIENTRY construct_particleSystem(const char* name,bool master) { return new particleSystem(name,master); }

struct TextureInfo
{
    pfTexture* pTexture;
    ygString filename;
    int row, column;
    int indexstart, indexend;
    bool loop;
    float cycle;
};

struct _particleSystemPrivateData
{
    Particle* m_Particles;
    string m_strChainSystem;
    particleSystem* m_pChainSystem;

    pfVec3   m_vPrevPosition;
    pfVec3   m_vPosition;

    int      m_nMaxParticle;
  
    float    m_fWidth, m_fHeight, m_fDepth;

    bool     m_bFloorParticle;
    bool     m_bCollideChain;
    pfVec3   m_vScale;
    int      m_nMaxParticleAlive;
    int      m_nParticleAlive;
    bool     m_bEmitEnable;
    pfVec3   m_vEmitPos, m_vEmitPosStart, m_vEmitPosEnd;
    float    m_fEmitRate;
    float    m_fLifeMin, m_fLifeMax;
    pfVec4   m_ColorStart, m_ColorStart2;
    pfVec4   m_ColorEnd, m_ColorEnd2;
    float    m_fSize, m_fSizeEnd;
    float    m_fYawMin, m_fYawMax;
    float    m_fPitchMin, m_fPitchMax;
    float    m_fRollMin, m_fRollMax;
    pfVec3   m_vGravity;
    float    m_fSpeedMin, m_fSpeedMax;

    bool     m_bCollide;
    bool     m_bAttractive;
    pfVec3   m_vAttractor;
    float    m_fAfterimageTime;
    int      m_nAfterimageStep;
    float    m_fEmissionResidue, m_fInitialParticle;
    float    m_fAlpha, m_fAlphaEnd;
    int      m_nBlendType;
    int      m_dwParticlesAlive;
    
    TextureInfo textureinfo;

    float    m_fTime;
    float    m_fElapsedTime;

    ygString m_sFilename;
};
     

/*************************************************************************/
// Particle
/*************************************************************************/
Particle::Particle()
{
#ifdef WIN32    
    srand(time(NULL));
#endif    
    m_pParentSystem = NULL;
    
    m_Color.set(1.0f, 0.0f, 0.0f, 0.5f);
    m_fGravity = 980.0f / 30.0f;
    m_fAge = 0.0f;
    m_fLifetime = 0.0f;
    m_fSize = 1.0f;
    m_fAlpha = 0.0f;
    
    // local geometry data manipulated by Particle itself
    m_pVertices = (pfVec3 *) pfMalloc(4*sizeof(pfVec3),pfGetSharedArena());
    m_pNormal  = (pfVec3 *) pfMalloc(4*sizeof(pfVec3),pfGetSharedArena());
    m_pTCoord  = (pfVec2 *) pfMalloc(4*sizeof(pfVec2),pfGetSharedArena());
    m_pColor    = (pfVec4 *) pfMalloc(4*sizeof(pfVec4),pfGetSharedArena());     

    geode = new pfGeode;
    geoset = new pfGeoSet;
    gstate = new pfGeoState;
    dcs = new pfDCS;
	  uint isectMask;
	  isectMask &= ~YG_ISECT_WALLMASK;
    geode->setTravMask(PFTRAV_ISECT,isectMask,PFTRAV_SELF|PFTRAV_DESCEND,PF_SET);

}

Particle::~Particle()
{
    m_pParentSystem->pfnode()->removeChild(dcs);
    
    pfDelete(m_pVertices);
    pfDelete(m_pNormal);
    pfDelete(m_pTCoord);
    pfDelete(m_pColor);
}

void Particle::initGeometry(int iBlendType)
{
    // init geometry data
    m_pVertices[0].set( 0.5, 0, 0);
    m_pVertices[1].set( 0.5, 0, 1);
    m_pVertices[2].set(-0.5, 0, 1);
    m_pVertices[3].set(-0.5, 0, 0);
    
    m_pNormal[0].set(0.0, -1.0, 0.0);
    m_pNormal[1].set(0.0, -1.0, 0.0);
    m_pNormal[2].set(0.0, -1.0, 0.0);
    m_pNormal[3].set(0.0, -1.0, 0.0);
    
    m_pTCoord[0].set(1,0);
    m_pTCoord[1].set(1,1);
    m_pTCoord[2].set(0,1);
    m_pTCoord[3].set(0,0);
    
    m_pColor[0].set(1.0, 1.0, 1.0, 1.0);
    m_pColor[1].set(1.0, 1.0, 1.0, 1.0);
    m_pColor[2].set(1.0, 1.0, 1.0, 1.0);
    m_pColor[3].set(1.0, 1.0, 1.0, 1.0);
    
    //create the geometry set and add the geometry data to it
    geoset->setAttr(PFGS_COORD3, PFGS_PER_VERTEX, m_pVertices, NULL);
    geoset->setAttr(PFGS_NORMAL3, PFGS_PER_VERTEX, m_pNormal, NULL);
    geoset->setAttr(PFGS_TEXCOORD2, PFGS_PER_VERTEX, m_pTCoord, NULL);
    geoset->setAttr(PFGS_COLOR4, PFGS_PER_VERTEX, m_pColor, NULL);
    geoset->setPrimType(PFGS_QUADS);
    geoset->setNumPrims(1);
    geode->addGSet(geoset);

    // texture
    if (m_pTexture)
    {
        gstate->setAttr(PFSTATE_TEXTURE, m_pTexture);
        gstate->setMode(PFSTATE_ENTEXTURE,PF_ON);
        int comp;
        
        //get number of image components
        m_pTexture->getImage(NULL,&comp,NULL,NULL,NULL);

        //if alpha channel then set transparency
        if (comp == 4)
        {
            gstate->setMode(PFSTATE_TRANSPARENCY,PFTR_ON|PFTR_NO_OCCLUDE|PFTR_BLEND_ALPHA);
            geoset->setDrawBin(PFSORT_TRANSP_BIN);
            //gstate->setMode(PFSTATE_ALPHAFUNC, PFAF_NOTEQUAL);
            //pfGStateVal (gstate, PFSTATE_ALPHAREF, 0.0f);
        }
        else
        {
            gstate->setMode(PFSTATE_TRANSPARENCY,PFTR_OFF);
            geoset->setDrawBin(PFSORT_OPAQUE_BIN);
        }

        dcs->setTravMask(PFTRAV_DRAW, 0xffffffff, PFTRAV_SELF|PFTRAV_DESCEND, PF_SET);
    }

    //Create a new material
    pfMaterial *material = new pfMaterial;
    
    //Create a pfGeostate, add the material, and then assign it to our geometry
    gstate->setAttr(PFSTATE_FRONTMTL, material);
    gstate->setAttr(PFSTATE_BACKMTL,  material);
    gstate->setAttr(PFSTATE_TEXTURE, m_pTexture);
    gstate->setMode(PFSTATE_CULLFACE, PF_OFF);
    gstate->setMode(PFSTATE_ENTEXTURE, PF_ON);

    m_pTev = new pfTexEnv;

    if (iBlendType == 1)
        m_pTev->setMode(PFTE_MODULATE);  // PFTE_MODULATE, PFTE_BLEND, PFTE_DECAL, PFTE_ADD
    else if (iBlendType == 2)
        m_pTev->setMode(PFTE_BLEND);
    else if (iBlendType == 3)
        m_pTev->setMode(PFTE_DECAL);
    else if (iBlendType == 4)
        m_pTev->setMode(PFTE_ADD);

    
    gstate->setAttr(PFSTATE_TEXENV, m_pTev);
    geoset->setGState(gstate);
    axis.set(0,0,1);

    dcs->addChild(geode);
    m_pParentSystem->pfnode()->addChild(dcs);
}

void Particle::setBlend(int iBlendType)
{
    if (iBlendType == 1)
        m_pTev->setMode(PFTE_MODULATE);  // PFTE_MODULATE, PFTE_BLEND, PFTE_DECAL, PFTE_ADD
    else if (iBlendType == 2)
        m_pTev->setMode(PFTE_BLEND);
    else if (iBlendType == 3)
        m_pTev->setMode(PFTE_DECAL);
    else if (iBlendType == 4)
        m_pTev->setMode(PFTE_ADD);

}

void Particle::Restart()
{

    m_fAge = m_fLifetime = 0.0f;
    m_vPrevPosition = m_vPosition = pfVec3(0,0,0);
    dcs->setTravMask(PFTRAV_DRAW,0xffffffff,PFTRAV_SELF,PF_SET); // draw
    
}

bool Particle::Update(float fTimeDelta)
{

    if ( m_fAge + fTimeDelta >= m_fLifetime ) 
    {
        // do something here to hide this particle
        m_fAge = -1.0f;
        dcs->setTravMask(PFTRAV_DRAW,0,PFTRAV_SELF,PF_SET); // hide
        return false;
    }
    else
    {
        m_fAge += fTimeDelta;

        m_vPrevPosition = m_vPosition;
        m_vPosition += fTimeDelta * m_vVelocity;

        if ( m_pParentSystem->IsColliding() )
        {
            if ( m_vPosition[2] < 0.0f ) 
            {
                if((m_pParentSystem->p_->m_bCollideChain) && (m_pParentSystem->p_->m_pChainSystem))
                {
                    m_fAge = -1.0f;
                    m_vPosition[2] = 0.0f;
                    m_pParentSystem->p_->m_pChainSystem->p_->m_vEmitPos = m_vPosition;
                    m_pParentSystem->p_->m_pChainSystem->p_->m_fEmissionResidue += 
                                m_pParentSystem->p_->m_pChainSystem->p_->m_fInitialParticle;
                    m_pParentSystem->p_->m_pChainSystem->Update(0);
                }
                else if ( m_vVelocity[2] > -5.0f ) 
                {
                    m_vPosition[2] = 0.0f;
                    m_vVelocity[2] = 0.0f;
                }
                else
                {
                    m_vPosition[2] = m_vPrevPosition[2]; 
                    m_vVelocity[2] = -m_vVelocity[2]*0.9f;
                }
            }
        }

        m_vVelocity += ( m_pParentSystem->p_->m_vGravity * fTimeDelta);

        if ( m_pParentSystem->IsAttractive() )
        {
            pfVec3 AttractLocation;
            
            m_pParentSystem->GetLocation(AttractLocation);

            pfVec3 AttractDir = AttractLocation - m_vPosition; 
            AttractDir.normalize();
            m_vVelocity += AttractDir * 10.0f * fTimeDelta;
        
        }

        m_Color += m_ColorDelta * fTimeDelta;
        m_fSize += m_fSizeDelta * fTimeDelta;
        m_fRoll += m_fRollDelta * fTimeDelta;
        m_fAlpha += m_fAlphaDelta * fTimeDelta;
        
        // changing vertex color & alpha
        pfVec4 tempColor = m_Color;
        tempColor[3] = m_fAlpha;
        m_pColor[0] = m_pColor[1] = m_pColor[2] =    m_pColor[3] = tempColor;
    }

    return true;
}

void Particle::UpdateTransform()
{
    //find viewer in the world
    ygNode* viewer = ygWorld::world()->viewer();
    //if a viewer is found
    if (viewer) 
    {
        //get viewer position vector relative to this node
        pfVec3 viewerVector = viewer->origin(m_pParentSystem->parent()); 
        viewerVector.normalize();

        //get the cross product of the viewer vector and the rotation axis
        pfVec3 viewCross;
        viewCross.cross(axis,viewerVector);
        viewCross.normalize();

        //establish the object front as the negative Y-axis
        pfVec3 frontAxis(0,-1,0);
        pfVec3 frontCross;
        
        //get the cross product of the front axis and the rotation axis
        frontCross.cross(axis,frontAxis);
        frontCross.normalize();
        float adjacent = frontCross.dot(viewCross);
        pfVec3 thirdNormal;
        
        //get a normal towards the viewer by crossing this cross with the rotaion axis
        thirdNormal.cross(axis,frontCross);
        float opposite = thirdNormal.dot(viewCross);
        
        //get the rotation to the viewer vector
        float axisRot = pfArcTan2(opposite,adjacent);
        
        //create a matrix with this rotation
        pfMatrix matrix;
        matrix.makeRot(axisRot,axis[0],axis[1],axis[2]);
        matrix.makeTrans(m_vPosition[0], m_vPosition[1], m_vPosition[2]);
        dcs->setMat(matrix);
        
    }
}


/*************************************************************************/
// particleSystem Constructor
/*************************************************************************/
particleSystem::particleSystem(const char* name,bool master) : ygNode(name,master)
{
    setClassName("particleSystem", true);
    p_ = new struct _particleSystemPrivateData;
    p_->m_Particles       = NULL;
    
    // original fire setting
    p_->m_nMaxParticle        = MAXPARTICLENUM;
    p_->m_nMaxParticleAlive = 111;
    p_->m_vEmitPos.set(0,0,0);
    p_->m_vEmitPosStart.set(-0.33f,0.0f,0.5f);
    p_->m_vEmitPosEnd.set(0.33f,0.0f,0.5f);
    p_->m_fEmitRate = 70.0f;
    p_->m_fLifeMin = 1.0f;        p_->m_fLifeMax = 2.0f;
    p_->m_ColorStart.set(1.0f,1.0f,1.0f, 1.0f);
    p_->m_ColorStart2.set(1.0f,1.0f,1.0f,1.0f);
    p_->m_ColorEnd.set(1.0f,0.0f,0.0f, 1.0f);
    p_->m_ColorEnd2.set(0.0f,0.0f,0.392157f, 1.0f);
    p_->m_fAlpha = 0.8f;        p_->m_fAlphaEnd = 0.0f;
    p_->m_fSize = 0.66f;        p_->m_fSizeEnd    = 1.66f;
    p_->m_fYawMin = 0.0f;        p_->m_fYawMax = 360.0f;
    p_->m_fPitchMin = 90.0f;    p_->m_fPitchMax = 90.0f;
    p_->m_fRollMin = 0.0f;    p_->m_fRollMax = 0.0f;
    p_->m_vGravity.set(0.33f,0.0f,3.33f);
    p_->m_fSpeedMin = 0.133f;    p_->m_fSpeedMax = 0.5f;
    p_->m_bCollide    = 0;
    p_->m_bAttractive = 0;
    p_->m_vAttractor.set(0.0f,0.0f,0.0f);
    
    p_->textureinfo.filename = "fire.rgb";

    p_->m_fAfterimageTime = 0.0f;
    p_->m_nAfterimageStep = 0;
    p_->m_fEmissionResidue = 0.0f;
    p_->m_fInitialParticle = p_->m_fEmissionResidue;
    p_->m_nBlendType = 2;
    p_->m_vScale.set(1.0f,1.0f,2.0f);
    p_->m_bFloorParticle = 0;
    p_->m_bCollideChain = 0;
    p_->m_strChainSystem = "";

    p_->m_bEmitEnable = true;
    p_->m_fWidth = p_->m_fHeight = p_->m_fDepth = 10;
    p_->m_vPrevPosition.set(0, 0, 0);
    p_->m_vPosition.set(0, 0, 0);

    p_->textureinfo.column = 8;
    p_->textureinfo.row = 4;
    p_->textureinfo.cycle = 1.5;
    p_->textureinfo.indexstart = 1;
    p_->textureinfo.indexend = 15;
    p_->textureinfo.loop = 1;

    addNetMessage("file",&p_->m_sFilename,YG_NET_FILE);
    addNetMessage("emitpos",&p_->m_vEmitPos,YG_NET_VEC3);
    addNetMessage("emit",&p_->m_bEmitEnable,YG_NET_BOOL);
    addNetMessage("blend",&p_->m_nBlendType,YG_NET_INT);
    addNetMessageMap("blend","modulate",1);
    addNetMessageMap("blend","blend",2);
    addNetMessageMap("blend","decal",3);
    addNetMessageMap("blend","add",4);
    addNetMessage("emitstart",&p_->m_vEmitPosStart,YG_NET_VEC3);
    addNetMessage("emitend",&p_->m_vEmitPosEnd,YG_NET_VEC3);
    addNetMessage("emitrate",&p_->m_fEmitRate,YG_NET_FLOAT);
    addNetMessage("maxParticles",&p_->m_nMaxParticleAlive,YG_NET_INT);
    addNetMessage("lifeMin",&p_->m_fLifeMin,YG_NET_FLOAT);
    addNetMessage("lifeMax",&p_->m_fLifeMax,YG_NET_FLOAT);
    addNetMessage("speedMin",&p_->m_fSpeedMin,YG_NET_FLOAT);
    addNetMessage("speedMax",&p_->m_fSpeedMax,YG_NET_FLOAT);
    addNetMessage("size",&p_->m_fSize,YG_NET_FLOAT);
    addNetMessage("sizeend",&p_->m_fSizeEnd,YG_NET_FLOAT);

    // initialize system geometry data
    InitSystem();
}

particleSystem::~particleSystem()
{
    if (p_->m_pChainSystem)
        delete p_->m_pChainSystem;

    delete[] p_->m_Particles;
}

void particleSystem::InitSystem()
{
    if (p_->m_Particles)
        delete [] p_->m_Particles;
        
    p_->m_Particles       = new Particle[p_->m_nMaxParticle];

    ygString texturefullPath;
    ygFindFileInYgPath(p_->textureinfo.filename,texturefullPath);
    
    for (int i=0; i < p_->m_nMaxParticle; i++)
    {
        p_->m_Particles[i].m_pTexture = ygPFTextureCache::textureReference(texturefullPath.c_str());
        p_->m_Particles[i].m_pParentSystem = this;
        p_->m_Particles[i].initGeometry(1);
    }
    
    p_->m_pChainSystem        = NULL;

    // reset timer
    p_->m_fTime = pfGetTime();
    p_->m_fElapsedTime = 0.0f;    
}

void particleSystem::reset(void)
{
    ResetPartlceArray();
    ygNode::reset();
}

void particleSystem::app(void)
{
    // update elapsed time
    float curTime = pfGetTime();
    
    p_->m_fElapsedTime = curTime - p_->m_fTime;
    p_->m_fTime = curTime;

    Update(p_->m_fElapsedTime);
    
    ygNode::app();
}

bool particleSystem::IsColliding()
{
    return p_->m_bCollide;
}

bool particleSystem::IsAttractive()
{
    return p_->m_bAttractive;
}

bool particleSystem::Update(float fTime)
{
    p_->m_dwParticlesAlive = 0;

    float theta, s, c;
    int textureindex;
    float agemod;
    pfVec3 temp;
    pfVec3* v;    // temporary pointer of Particle vertex position & normal
    pfVec4*    vc;    // temporary pointer of Particle vertex color
    pfVec2* vt;    // temporary pointer of Particle vertex texture coord
    
    int i = 0; // for loop

    for ( i=0; i < p_->m_nMaxParticle; i++ )
    {
        if (p_->m_Particles[i].m_fAge >= 0.0f )
            if ( p_->m_Particles[i].Update(fTime))
            {
                if(p_->textureinfo.indexend - p_->textureinfo.indexstart != 0)
                {
                    // texture coordinate
                    agemod = (p_->m_Particles[i].m_fAge + p_->m_Particles[i].m_fGap) /p_->textureinfo.cycle;
                    agemod = agemod - (float)floor(agemod);
                    textureindex = (int)(agemod * (p_->textureinfo.indexend - p_->textureinfo.indexstart) + p_->textureinfo.indexstart + 0.5f);
                    vt = p_->m_Particles[i].m_pTCoord;
                    
                    vt[2].set( (float)((textureindex-1) % p_->textureinfo.column) /p_->textureinfo.column, 
                                1 - (float)((textureindex-1) / p_->textureinfo.column) / p_->textureinfo.row);
                    vt[3].set( (float)((textureindex-1) % p_->textureinfo.column) /p_->textureinfo.column, 
                                1 - (float)((textureindex-1) /p_-> textureinfo.column + 1) / p_->textureinfo.row);
                    vt[1].set( (float)((textureindex-1) % p_->textureinfo.column + 1) /p_->textureinfo.column,
                                1 - (float)((textureindex-1) / p_->textureinfo.column) / p_->textureinfo.row);
                    vt[0].set( (float)((textureindex-1) % p_->textureinfo.column + 1) /p_->textureinfo.column,
                                1 - (float)((textureindex-1) / p_->textureinfo.column + 1) / p_->textureinfo.row);        
                }

                if((p_->m_Particles[i].m_fRollDelta != 0.0f) || (p_->m_Particles[i].m_fSizeDelta != 0.0f))
                {
										theta = PT_PI / 4 + p_->m_Particles[i].m_fRoll;
                    s = sinf(theta) * p_->m_Particles[i].m_fSize;    c = cosf(theta) * p_->m_Particles[i].m_fSize;
                    
                    v = p_->m_Particles[i].m_pVertices;
                    
                    if(p_->m_bFloorParticle)
                    {
                        v[3].set(-c, -s, 0);
                        v[2].set(-s, c, 0);
                        v[0].set(s, -c, 0);
                        v[1].set(c, s, 0);
                        temp = (v[3] + v[2]) / 2.0f;
                        v[3] = (v[3] - temp) * p_->m_vScale[1] + temp;
                        v[2] = (v[2] - temp) * p_->m_vScale[1] + temp;
                        temp = (v[0] + v[1]) / 2.0f;
                        v[0] = (v[0] - temp) * p_->m_vScale[1] + temp;
                        v[1] = (v[1] - temp) * p_->m_vScale[1] + temp;
                    }
                    else
                    {
                        
                        v[3].set(-c, 0, -s + ROOT2INV);
                        v[2].set(-s, 0, c + ROOT2INV);
                        v[0].set(s, 0, -c + ROOT2INV);
                        v[1].set(c, 0, s + ROOT2INV);
                        temp = (v[3] + v[2]) / 2.0f;
                        v[3] = (v[3] - temp) * p_->m_vScale[2] + temp;
                        v[2] = (v[2] - temp) * p_->m_vScale[2] + temp;
                        temp = (v[0] + v[1]) / 2.0f;
                        v[0] = (v[0] - temp) * p_->m_vScale[2] + temp;
                        v[1] = (v[1] - temp) * p_->m_vScale[2] + temp;
                    }
                    temp = (v[3] + v[0]) / 2.0f;
                    v[3] = (v[3] - temp) * p_->m_vScale[0] + temp;
                    v[0] = (v[0] - temp) * p_->m_vScale[0] + temp;
                    temp = (v[2] + v[1]) / 2.0f;
                    v[2] = (v[2] - temp) * p_->m_vScale[0] + temp;
                    v[1] = (v[1] - temp) * p_->m_vScale[0] + temp;
                }
                p_->m_dwParticlesAlive++;
            }
    }
    if (p_->m_nMaxParticleAlive < (int)p_->m_dwParticlesAlive)
        p_->m_nMaxParticleAlive = p_->m_dwParticlesAlive;

    float fParticlesNeeded = p_->m_fEmitRate * fTime + p_->m_fEmissionResidue;
    
    int dwParticlesCreated = (unsigned int)fParticlesNeeded;
    
    if ( p_->m_bEmitEnable )
    {
        p_->m_fEmissionResidue = fParticlesNeeded - dwParticlesCreated;
    }
    else
    {
        p_->m_fEmissionResidue = fParticlesNeeded;
        dwParticlesCreated = 0;
    }
    
    float fRand;
    if ( dwParticlesCreated > 0 )
    {
        for ( i=0; i < p_->m_nMaxParticle; i++ )
        {
            if ( !dwParticlesCreated )
                break;

            if ( p_->m_Particles[i].m_fAge < 0.0f ) 
            {
                // do something here to draw object
                p_->m_Particles[i].dcs->setTravMask(PFTRAV_DRAW,0xffffffff,PFTRAV_SELF,PF_SET); // draw
                
                p_->m_Particles[i].m_fAge = 0.0f;
                p_->m_Particles[i].m_fLifetime = RandomIn(p_->m_fLifeMin, p_->m_fLifeMax);

                fRand = RandomIn(0, 1);
                p_->m_Particles[i].m_Color = p_->m_ColorStart + fRand * (p_->m_ColorStart2 - p_->m_ColorStart);
                
                pfVec4 tempColor;
                fRand = RandomIn(0, 1);
                tempColor =  p_->m_ColorEnd + fRand * (p_->m_ColorEnd2 - p_->m_ColorEnd) ;
                tempColor[3] = 0;
                p_->m_Particles[i].m_ColorDelta = (tempColor - p_->m_Particles[i].m_Color) / p_->m_Particles[i].m_fLifetime ;

                p_->m_Particles[i].m_fSize = p_->m_fSize;
                p_->m_Particles[i].m_fSizeDelta = (p_->m_fSizeEnd - p_->m_fSize) / p_->m_Particles[i].m_fLifetime;
                
                if(p_->textureinfo.loop)
                    p_->m_Particles[i].m_fGap = RandomIn(0, p_->textureinfo.cycle);
                else
                    p_->m_Particles[i].m_fGap = 0;
                
                if (p_->m_fRollMax != p_->m_fRollMin) 
                    p_->m_Particles[i].m_fRoll = RandomIn(0, 360);
                else
                    p_->m_Particles[i].m_fRoll = 0;

                p_->m_Particles[i].m_fRollDelta = RandomIn(p_->m_fRollMax, p_->m_fRollMin);

                p_->m_Particles[i].m_fAlpha = p_->m_fAlpha;
                p_->m_Particles[i].m_fAlphaDelta = (p_->m_fAlphaEnd - p_->m_fAlpha) / p_->m_Particles[i].m_fLifetime;

                pfVec3 tempVelocity = (p_->m_vPosition - p_->m_vPrevPosition)/fTime;
                
                if(fTime == 0)
                    tempVelocity.set(0, 0, 0);
                    
                pfVec3 LocalPos = pfVec3(RandomIn(p_->m_vEmitPosStart[0], p_->m_vEmitPosEnd[0]), 
                                                        RandomIn(p_->m_vEmitPosStart[1], p_->m_vEmitPosEnd[1]), 
                                                        RandomIn(p_->m_vEmitPosStart[2], p_->m_vEmitPosEnd[2]));

                p_->m_Particles[i].m_vPosition = p_->m_vPrevPosition + LocalPos + tempVelocity * RandomIn(0, 1) * fTime + p_->m_vEmitPos;
                p_->m_Particles[i].m_vPrevPosition = p_->m_vPosition;

				    float RandomYaw = RandomIn(p_->m_fYawMin, p_->m_fYawMax);
				    float RandomPitch = RandomIn(p_->m_fPitchMin, p_->m_fPitchMax);
				    
				    p_->m_Particles[i].m_vVelocity.set( sinf(RandomPitch) * cosf(RandomYaw), 
																		sinf(RandomPitch) * sinf(RandomYaw),
																		cosf( RandomPitch ));

                p_->m_Particles[i].m_vVelocity *= RandomIn(p_->m_fSpeedMin, p_->m_fSpeedMax);
                
                vc = p_->m_Particles[i].m_pColor;
                
                tempColor = pfVec4(p_->m_Particles[i].m_Color[0], 
                                                p_->m_Particles[i].m_Color[1], 
                                                p_->m_Particles[i].m_Color[2], 
                                                0.0f/*p_->m_Particles[i].m_Color[3]*/);
                vc[0] = vc[1] = vc[2] =    vc[3] = tempColor;

                // texture coordinate
                agemod = (p_->m_Particles[i].m_fAge + p_->m_Particles[i].m_fGap) /p_->textureinfo.cycle;
                agemod = agemod - (float)floor(agemod);
                textureindex = (int)(agemod * (p_->textureinfo.indexend - p_->textureinfo.indexstart) + p_->textureinfo.indexstart + 0.5f);
                
                vt = p_->m_Particles[i].m_pTCoord;
                            
                vt[2].set( (float)((textureindex-1) % p_->textureinfo.column) /p_->textureinfo.column,
                         1 - (float)((textureindex-1) / p_->textureinfo.column) / p_->textureinfo.row);
                vt[3].set( (float)((textureindex-1) % p_->textureinfo.column) /p_->textureinfo.column, 
                        1 - (float)((textureindex-1) / p_->textureinfo.column + 1) / p_->textureinfo.row );
                vt[1].set( (float)((textureindex-1) % p_->textureinfo.column + 1) /p_->textureinfo.column,
                        1 - (float)((textureindex-1) / p_->textureinfo.column) / p_->textureinfo.row );
                vt[0].set( (float)((textureindex-1) % p_->textureinfo.column + 1) /p_->textureinfo.column, 
                        1 - (float)((textureindex-1) / p_->textureinfo.column + 1) / p_->textureinfo.row );

								theta = PT_PI / 4 + p_->m_Particles[i].m_fRoll;
                s = sinf(theta) * p_->m_Particles[i].m_fSize;    c = cosf(theta) * p_->m_Particles[i].m_fSize;
                
                v = p_->m_Particles[i].m_pVertices;

                if(p_->m_bFloorParticle)
                {
                    v[3].set(-c, -s, 0);
                    v[2].set(-s, c, 0);
                    v[0].set(s, -c, 0);
                    v[1].set(c, s, 0);
                    temp = (v[3] + v[2]) / 2.0f;
                    v[3] = (v[3] - temp) * p_->m_vScale[1] + temp;
                    v[2] = (v[2] - temp) * p_->m_vScale[1] + temp;
                    temp = (v[0] + v[1]) / 2.0f;
                    v[0] = (v[0] - temp) * p_->m_vScale[1] + temp;
                    v[1] = (v[1] - temp) * p_->m_vScale[1] + temp;
                }
                else
                {
                    
                    v[3].set(-c, 0, -s + ROOT2INV);
                    v[2].set(-s, 0, c + ROOT2INV);
                    v[0].set(s, 0, -c + ROOT2INV);
                    v[1].set(c, 0, s + ROOT2INV);
                    temp = (v[3] + v[2]) / 2.0f;
                    v[3] = (v[3] - temp) * p_->m_vScale[2] + temp;
                    v[2] = (v[2] - temp) * p_->m_vScale[2] + temp;
                    temp = (v[0] + v[1]) / 2.0f;
                    v[0] = (v[0] - temp) * p_->m_vScale[2] + temp;
                    v[1] = (v[1] - temp) * p_->m_vScale[2] + temp;
                }
                temp = (v[3] + v[0]) / 2.0f;
                v[3] = (v[3] - temp) * p_->m_vScale[0] + temp;
                v[0] = (v[0] - temp) * p_->m_vScale[0] + temp;
                temp = (v[2] + v[1]) / 2.0f;
                v[2] = (v[2] - temp) * p_->m_vScale[0] + temp;
                v[1] = (v[1] - temp) * p_->m_vScale[0] + temp;
                dwParticlesCreated--;
            }
        }
    }

    p_->m_vPrevPosition = p_->m_vPosition;    

    // apply billboard transform
    for ( i=0; i < p_->m_nMaxParticle; i++ )
    {
        if (p_->m_Particles[i].m_fAge >= 0.0f )
        {
            p_->m_Particles[i].UpdateTransform();
        }
    }
    
    if(p_->m_pChainSystem)
        p_->m_pChainSystem->Update(fTime);

    return true;
}

void particleSystem::Move(pfVec3 vPosition)
{

}

int particleSystem::NumParticles()
{
    return 0;
}

void particleSystem::message(const ygMessage& msg)
{
    //load particle system editor file
    if (msg == "file")
    {
        if (p_->m_sFilename.size() > 0)
        {
            if(LoadSFXFile(p_->m_sFilename))
                ResetPartlceArray();
        }
    }
    else if (msg == "restart")
    {
        Restart();
    }
    //this message is for moving emit position
    else if (msg == "emitpos")        
    {
        return;
    }
    //minimum emit start location
    else if (msg == "emitstart")
    {
        return;
    }
    //maximum emit start location
    else if (msg == "emitend")
    {
        return;
    }
    //set emit rate
    else if (msg == "emitrate")
    {
        return;
    }
    //turn emit on/off
    else if (msg == "emit")
    {
        return;
    }
    //set blend type
    else if (msg == "blend")
    {
        setBlend(p_->m_nBlendType);
    }
    //set max number of alive particles
    else if (msg == "maxParticles")
    {
        return;
    }
    //set minimum particle life
    else if (msg == "lifeMin")
    {
        return;
    }
    //set maximum particle life
    else if (msg == "lifeMax")
    {
        return;
    }
    //set initial particle size
    else if (msg == "size")
    {
        return;
    }
    //set end particle size
    else if (msg == "sizeend")
    {
        return;
    }
    //set minimum particle speed
    else if (msg == "speedMin")
    {
        return;
    }
    //set maximum particle speed
    else if (msg == "speedMax")
    {
        return;
    }
    else
        ygNode::message(msg);
}

static float GetFloatToken(char* str)
{
    char strTmp[MAX_LENGTH];
    int i = 0;
    
    for( i=0; i<=strlen(str) && str[i] != ',' && str[i] != '\n'; i++)
        strTmp[i] = str[i];
        
    strcpy(str, str+i+1);
    return atof(strTmp);
}

static pfVec3 GetVec3Token(char* str)
{
    float x, y, z;
    x = CM2Feet(GetFloatToken(str));
    y = CM2Feet(GetFloatToken(str));
    z = CM2Feet(GetFloatToken(str));
    return pfVec3(x, z, y);
}

void GetStringToken(char* str)
{
    char strTmp1[MAX_LENGTH];
    
    int i;
    for (i=0; i<MAX_LENGTH; i++)
    {    
        strTmp1[i]=0;
    }

    sscanf(str, "%s", &strTmp1);        
    strcpy(str, strTmp1);
}

void GetToken(FILE *ar, char* str)
{
    fgets(str, MAX_LENGTH, ar);
    // before the ':'
    int i;
    for (i=0; str[i] != ':' && i <= strlen(str); i++)
    {
        // skip
         ;    
    }
    strcpy(str, str+i+1);
    return;
}

bool particleSystem::LoadSFXFile(const ygString& file)
{
    // declare variables
    char str[MAX_LENGTH];
    FILE *infile;

    ygString fullPath;
    if (!ygFindFileInYgPath(file,fullPath))
        return false;
    
    if ((infile = fopen( fullPath.c_str(), "rt" )) == NULL)
    {
        printf("Cannot open input file - %s\n", fullPath.c_str() );
        return false;
    }

    // read first two line: spec version and empty line
    fgets(str, MAX_LENGTH, infile);
    fgets(str, MAX_LENGTH, infile);


    // modified for unit change(from DX to PF)
    // read and set values
     GetToken(infile, str);    p_->m_nMaxParticleAlive = atoi(str);
                                   p_->m_nMaxParticle = (int)(p_->m_nMaxParticleAlive * 1.1);
    GetToken(infile, str);    p_->m_vEmitPosStart = GetVec3Token(str);
    GetToken(infile, str);    p_->m_vEmitPosEnd = GetVec3Token(str);
    GetToken(infile, str);    p_->m_fEmitRate = GetFloatToken(str);
    GetToken(infile, str);    p_->m_fLifeMin = GetFloatToken(str);        
                                     p_->m_fLifeMax = GetFloatToken(str);
    GetToken(infile, str);    p_->m_ColorStart[0] = GetFloatToken(str);    
                                     p_->m_ColorStart[1] = GetFloatToken(str);    
                                     p_->m_ColorStart[2] = GetFloatToken(str);
    GetToken(infile, str);    p_->m_ColorStart2[0] = GetFloatToken(str);    
                                     p_->m_ColorStart2[1] = GetFloatToken(str);    
                                     p_->m_ColorStart2[2] = GetFloatToken(str);
    GetToken(infile, str);    p_->m_ColorEnd[0] = GetFloatToken(str);        
                                     p_->m_ColorEnd[1] = GetFloatToken(str);        
                                     p_->m_ColorEnd[2] = GetFloatToken(str);
    GetToken(infile, str);    p_->m_ColorEnd2[0] = GetFloatToken(str);    
                                     p_->m_ColorEnd2[1] = GetFloatToken(str);    
                                     p_->m_ColorEnd2[2] = GetFloatToken(str);
    GetToken(infile, str);    p_->m_fAlpha = GetFloatToken(str);        
                                     p_->m_fAlphaEnd = GetFloatToken(str);
    GetToken(infile, str);    p_->m_fSize = CM2Feet(GetFloatToken(str));        
                                     p_->m_fSizeEnd    = CM2Feet(GetFloatToken(str));
	GetToken(infile, str);	p_->m_fYawMin = GetFloatToken(str);
									p_->m_fYawMax = GetFloatToken(str);
									p_->m_fYawMin = Degree2Radian(p_->m_fYawMin);
									p_->m_fYawMax = Degree2Radian(p_->m_fYawMax);
	GetToken(infile, str);	p_->m_fPitchMin = GetFloatToken(str);	
									p_->m_fPitchMax = GetFloatToken(str);
									p_->m_fPitchMin = Degree2Radian(p_->m_fPitchMin);
									p_->m_fPitchMax = Degree2Radian(p_->m_fPitchMax);
	GetToken(infile, str);	p_->m_fRollMin = GetFloatToken(str);	
									p_->m_fRollMax = GetFloatToken(str);
									p_->m_fRollMin = Degree2Radian(p_->m_fRollMin);
									p_->m_fRollMax = Degree2Radian(p_->m_fRollMax);
    GetToken(infile, str);    p_->m_vGravity = GetVec3Token(str);
    GetToken(infile, str);    p_->m_fSpeedMin = CM2Feet(GetFloatToken(str));    
                                     p_->m_fSpeedMax = CM2Feet(GetFloatToken(str));
    GetToken(infile, str);    p_->m_bCollide    = atoi(str);
    GetToken(infile, str);    p_->m_bAttractive = atoi(str);
    GetToken(infile, str);    p_->m_vAttractor = GetVec3Token(str);
    
    GetToken(infile, str);    GetStringToken(str); p_->textureinfo.filename = str;
    GetToken(infile, str);    p_->m_fAfterimageTime = GetFloatToken(str);
    GetToken(infile, str);    p_->m_nAfterimageStep = atoi(str);
    GetToken(infile, str);    p_->m_fEmissionResidue = GetFloatToken(str);
    p_->m_fInitialParticle = p_->m_fEmissionResidue;
    GetToken(infile, str);    p_->m_nBlendType = atoi(str);
    GetToken(infile, str);    p_->m_vScale = GetVec3Token(str) * 30;
    GetToken(infile, str);    p_->m_bFloorParticle = atoi(str);
    GetToken(infile, str);    p_->m_bCollideChain = atoi(str);
    GetToken(infile, str);    GetStringToken(str); p_->m_strChainSystem = str;

    GetToken(infile, str);    p_->textureinfo.column = atoi(str);
    GetToken(infile, str);    p_->textureinfo.row = atoi(str);
    GetToken(infile, str);    p_->textureinfo.cycle = GetFloatToken(str);
    GetToken(infile, str);    p_->textureinfo.indexstart = atoi(str);
    GetToken(infile, str);    p_->textureinfo.indexend = atoi(str);
    GetToken(infile, str);    p_->textureinfo.loop = atoi(str);
     
    // close the infile
    fclose(infile);
    
     return true;
}

void particleSystem::ResetPartlceArray()
{
    if (p_->m_Particles)
    {    
        delete [] p_->m_Particles;
    }
        
    p_->m_Particles       = new Particle[p_->m_nMaxParticle];

    ygString texturefullPath;
    ygFindFileInYgPath(p_->textureinfo.filename,texturefullPath);
    
    for (int i=0; i < p_->m_nMaxParticle; i++)
    {
        p_->m_Particles[i].m_pTexture = ygPFTextureCache::textureReference(texturefullPath.c_str());
        p_->m_Particles[i].m_pParentSystem = this;
        p_->m_Particles[i].initGeometry(p_->m_nBlendType);
    }
    
    if(p_->m_pChainSystem)
    {
        delete p_->m_pChainSystem;
    }
    
    // enable emit
    p_->m_bEmitEnable = true;
    
    // reset timer
    p_->m_fTime = pfGetTime();
    p_->m_fElapsedTime = 0.0f;    
    
    if ( p_->m_fEmitRate == 0)
    {    
        p_->m_fEmissionResidue = p_->m_fInitialParticle;
    }
}

void particleSystem::setBlend(int iBlendType)
{
    for (int i=0; i < p_->m_nMaxParticle; i++)
    {
        p_->m_Particles[i].setBlend(iBlendType);
    }

}

void particleSystem::Restart()
{
    p_->m_vEmitPos.set(0,0,0);
    p_->m_vPrevPosition.set(0, 0, 0);
    p_->m_vPosition.set(0, 0, 0);
    p_->m_bEmitEnable = true;
    
    p_->m_fEmissionResidue = p_->m_fInitialParticle;
    
    for (int i=0; i < p_->m_nMaxParticle; i++)
    {
        p_->m_Particles[i].Restart();
    }
    
    // reset timer
    p_->m_fTime = pfGetTime();
    p_->m_fElapsedTime = 0.0f;    

}

void particleSystem::LoadChainSystem(string &str)
{
    if(!p_->m_pChainSystem)
        p_->m_pChainSystem = new particleSystem("chain");

    //open file and read spec
    p_->m_pChainSystem->LoadSFXFile(str.c_str());
    
    p_->m_strChainSystem = str;
}

void particleSystem::KillThemAll()
{
        for (int i=0; i < p_->m_nMaxParticle; i++ )
            p_->m_Particles[i].m_fAge = -1.0f;
}

void particleSystem::GetLocation(pfVec3& vLocation)
{

}

