// Description: an interpolate node with transformed position and orientation values
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> the array size is 6 with the following values
//<ul>
//<li> value0 is the X translation
//<li> value1 is the Y translation
//<li> value2 is the Z translation
//<li> value3 is the X orientation
//<li> value4 is the Y orientation
//<li> value5 is the Z orientation
//</ul>
//<li> this node is useful for constraining and reporting transform position and orientation
//<li> this node is useful for grabbing and moving when used in conjunction with a <a href="wandTrigger.html">wandTrigger</a> node
//</ul>
//
// Category: Math
// Author: Alex Hill
//           06/01/02
// Revision: 07/01/06 Alex Hill - added logic to switch to unreliable messages when possible
//
#include <Performer/pf/pfDCS.h>
#include <Performer/pf/pfGeode.h>
#include <Performer/pr/pfGeoSet.h>
#include <Performer/pr/pfMaterial.h>
//#include <Performer/pfdu.h>
#include <ygNodeDB.h>
#include <ygNetMessages.h>
#include <ygWorld.h>
#include "valuator.h"
//#include "ygUtil.h"

extern "C" ygNode* APIENTRY construct_valuator(const char* name,bool master) { return new valuator(name,master); }
char* valuator_dependencies = "YG DSO Dependencies: interpolate";

#define COORD_NETKEY "x"

#define CONE_BASE 0.0
#define CONE_TOP 0.1
#define CONE_RADIUS 0.1
#define CONE_LENGTH 0.070711 
#define AXIS_WIDTH 0.01

#define CHANGE_THRESHOLD 3
#define CHANGE_INITIALIZED -1

valuator::valuator(const char* name,bool master) : interpolate(name,master)
    {
    setClassName("valuator",true);
    dcs = new pfDCS;
    setPfNode(dcs);
    netCoord.xyz.set(0,0,0);
    netCoord.hpr.set(0,0,0);
    lastNetCoord = netCoord;
    node = NULL;
    position.set(0,0,0);
    orientation.set(0,0,0);
    addNetMessage("grab",&nodeName,YG_NET_NODE);
    addNetMessage("position",&position,YG_NET_VEC3);
    dontSendNetMessage("position");
    addNetMessage("orientation",&orientation,YG_NET_VEC3);
    dontSendNetMessage("orientation");
    updateInterval = 0;
    if (getenv("YG_NET_TRANSFORM_INTERVAL"))
        updateInterval = atof(getenv("YG_NET_TRANSFORM_INTERVAL"));
    lastNetUpdate = 0;
    addNetMessage("updateinterval",&updateInterval,YG_NET_FLOAT);
    //create a simple coordinate network Message
    addNetMessage(COORD_NETKEY, &netCoord, YG_NET_COORD);
    //dont include in documentation
    dontSaveNetMessage(COORD_NETKEY);
    reliableMessage(COORD_NETKEY);
    }

valuator::~valuator(void)
    {
    }


void valuator::reset(void)
    {
    //if net master then 
    if (isNetMaster())
        {
        //reset matrix to identity
        deltaMatrix.makeIdent();
        //reset node pointer
        node = NULL;
        //reset grab flag
        grab = false;
        //set position to (0,0,0)
        netCoord.xyz.set(0,0,0);
        //set orientation to (0,0,0)
        netCoord.hpr.set(0,0,0);
        //set matrix to identity
        pfMatrix mat;
        mat.makeIdent();
        setMatrix(mat);
        resetNetMessage("grab");
        //set update interval to 0.0
        resetNetMessage("updateinterval");
        //set last update time to 0.0
        lastNetUpdate = 0;
        netCoordAuto = true;
        netCoordUnchanged = CHANGE_INITIALIZED;
        reliableMessage(COORD_NETKEY);
        }
    interpolate::reset();
    setValueSize(6);
    }


void valuator::message(const ygMessage& msg)
    {
    //toggle the grab/drop state with respect to a node
    if (msg == "toggle")
        {
        if (msg.numArgs() == 1)
            {
            if (nodeName == msg.stringArg(0) && node)
                {
                node = NULL;
                nodeName.clear();
                }
            else
                {
                nodeName = msg.stringArg(0);
                node = ygNodeDB::find(nodeName);
                if (node)
                    {
                    grab = true;
                    ygString args("node=");
                    args += node->name();
                    eventOccurred("grab",args);
                    }
                else
                    msg.error(name(),"(node not found)");
                }
            netMessageChanged("grab");
            }
        else
            msg.error(name(),"(wrong number of arguments)");
            }
    //grab a node and move with repect to that node
    else if (msg == "grab")
        {
        if (nodeName.length() > 0)
            {
            node = ygNodeDB::find(nodeName);
            if (node)
                {
                grab = true;
                ygString args("node=");
                args += node->name();
                eventOccurred("grab",args);
                }
            else
                msg.error(name(),"(node not found)");
            }
        }
    //drop connection with a node or any node grabbed
    else if (msg == "drop")
        {
        if (msg.numArgs() < 2)
            {
            if (msg.numArgs() == 1)
                {
                ygString dropName = msg.stringArg(0);
                if (dropName == nodeName)
                    {
                    node = NULL;
                    nodeName.clear();
                    netMessageChanged("grab");
                    }
                }
            else
                {
                node = NULL;
                nodeName.clear();
                netMessageChanged("grab");
                }
            }
        else
            msg.error(name(),"(wrong number of arguments)");
        }
    //set the translational component
    else if (msg == "position")
        {
        setValueSize(6);
        for (int i=0; i<3; i++)
            values[i] = startValues[i]+position[i]*(endValues[i]-startValues[i]);
        }
    //set the orientation component
    else if (msg == "orientation")
        {
        setValueSize(6);
        for (int i=3; i<6; i++)
            values[i] = startValues[i]+orientation[i-3]*(endValues[i]-startValues[i]);
        }
    else if (msg == COORD_NETKEY)
        {
        setPosition(netCoord.xyz);
        setOrientation(netCoord.hpr[1],netCoord.hpr[2],netCoord.hpr[0]);
        }
    //set the Message update interval for remote users
    else if (msg == "updateInterval")
        {
        return;
        }
    else
        interpolate::message(msg);
    }

void valuator::app(void)
    {
    //if net master
    if (isNetMaster())
        {
        calculateTransform();
        value::app();
        updateTransform();
        }
    //adjust debug display
    if (debugDisplay())
        {
        pfMatrix mat;
        getMatrix(mat);
        pfVec3 pos(0,0,0);
        pos.xformPt(pos,mat);
        pfVec3 vec;
        pfSphere sphere;
        sphere.radius = 0;
        sphere.center.set(0,0,0);
        float radius = 0;
        pfNode* debug = debugNode();
        for (int i=0; i<pfnode()->getNumChildren(); i++)
            {
            pfNode* child = pfnode()->getChild(i);
            if (child != debug)
                {
                child->getBound(&sphere);
                sphere.center.xformPt(sphere.center,mat);
                vec = sphere.center - pos;
                float currRadius = vec.length() + sphere.radius;
                if (currRadius > radius)
                    radius = currRadius;
                }
            }
        if (radius == 0.0)
            radius = 0.5;
        //if bounding sphere has changed then update debug
        if (radius != lastRadius)
            {
            setDebugCoords(radius);
            lastRadius = radius;
            }
        }
    }

void valuator::calculateTransform(void)
    {
    //if the node has been found
    if (node)
        {
        if (grab)
            {
            /*calculate the relative matrix between the node and this object
              using the following formula:
                                      -1
              delatMatrix = M      * M
                             self     node
            */
            pfMatrix node_matrix;
            node->getTransform(node_matrix);

            pfMatrix inverse_node_matrix;
            inverse_node_matrix.invertOrtho(node_matrix);

            pfMatrix self_matrix;
            getTransform(self_matrix);

            // Set this instance as "grabbed".
            deltaMatrix = self_matrix * inverse_node_matrix;
            grab = false;
            }
        /*move this matrix with respect to the node
                                      -1
           M = deltaMatrix * M     * M
                              node    parent
        */

        //get the node's transformation matrix
        pfMatrix node_matrix;
        node->getTransform(node_matrix);

        //get our object's transformation matrix relative to the world
        pfMatrix parent_matrix;
        getParentTransform(parent_matrix);
        parent_matrix.invertOrtho(parent_matrix);

        //set the new transformation and return
        pfMatrix self_matrix = deltaMatrix * node_matrix * parent_matrix;
        pfCoord coord;
        self_matrix.getOrthoCoord(&coord);
        //calculate interpolated array values
        if (size >= 3)
            {
            values[0] = startValues[0] + coord.xyz[0]*(endValues[0]-startValues[0]);
            values[1] = startValues[1] + coord.xyz[1]*(endValues[1]-startValues[1]);
            values[2] = startValues[2] + coord.xyz[2]*(endValues[2]-startValues[2]);
            }
        if (size >= 6)
            {
            values[3] = startValues[3] + coord.hpr[1]*(endValues[3]-startValues[3]);
            values[4] = startValues[4] + coord.hpr[2]*(endValues[4]-startValues[4]);
            values[5] = startValues[5] + coord.hpr[0]*(endValues[5]-startValues[5]);
            }
        }
    }

void valuator::updateTransform(void)
    {
    float x, y, z;
    float x_ang, y_ang, z_ang;
    if (size >= 3)
        {
        x = (values[0] - startValues[0])/(endValues[0]-startValues[0]);
        y = (values[1] - startValues[1])/(endValues[1]-startValues[1]);
        z = (values[2] - startValues[2])/(endValues[2]-startValues[2]);
        setPosition(x,y,z);
        }
    if (size >= 6)
        {
        x_ang = (values[3] - startValues[3])/(endValues[3]-startValues[3]);
        y_ang = (values[4] - startValues[4])/(endValues[4]-startValues[4]);
        z_ang = (values[5] - startValues[5])/(endValues[5]-startValues[5]);
        setOrientation(x_ang,y_ang,z_ang);
        }
    //if update interval has passed then
    if (updateInterval >= 0.0 && ygWorld::frameTime()-lastNetUpdate > updateInterval)
        {
        //if network coord has changed then
        if (netCoord.xyz != lastNetCoord.xyz ||
            netCoord.hpr != lastNetCoord.hpr)
            {
            //indicate message changed
            netMessageChanged(COORD_NETKEY);
            lastNetCoord = netCoord;
            lastNetUpdate = ygWorld::frameTime();
            //if unchanged counter is positive then
            if (netCoordUnchanged >= 0)
                {
                //if counter is over threshold then set unreliable
                if (netCoordAuto && netCoordUnchanged > CHANGE_THRESHOLD)
                    unreliableMessage(COORD_NETKEY);
                //set counter to zero
                netCoordUnchanged = 0;
                }
            //else initialize counter over threshold
            else 
                netCoordUnchanged = CHANGE_THRESHOLD+1;
            }
        //else if counter is under threshold then increment
        else if (netCoordUnchanged < CHANGE_THRESHOLD)
            {
            netCoordUnchanged++;
            netMessageChanged(COORD_NETKEY);
            }
        //else if counter is at threshold then set reliable
        else if (netCoordUnchanged == CHANGE_THRESHOLD)
            {
            netCoordUnchanged++;
            if (netCoordAuto)
                reliableMessage(COORD_NETKEY);
            }
        }
    }

/*****  DCS functions *****/

void valuator::getMatrix(pfMatrix& mat) const
    {
    dcs->getMat(mat);
    }

void valuator::setMatrix(pfMatrix& mat)
    {
    dcs->setMat(mat);
    }

void valuator::setPosition(float x,float y,float z)
    {
    position.set(x,y,z);
    netCoord.xyz.set(x,y,z);
    dcs->setTrans(x,y,z);
    }

void valuator::setOrientation(float x,float y,float z)
    {
    orientation.set(x,y,z);
    netCoord.hpr.set(z,x,y);
    dcs->setRot(z,x,y);
    }

void valuator::createDebugNode(void)
    {
    pfGeode* geode = new pfGeode;
    if (!geode)
        {
        cerr << "ERROR: ygTransform::createDebugGeode: failed to create pfGeode\n";
        return;
        }
    //create cone geoset
    pfGeoSet *gset = new pfGeoSet;
    int* lengths = (int *) pfMalloc(6*sizeof(int),pfGetSharedArena());
    gset->setPrimType(PFGS_TRISTRIPS);
    gset->setNumPrims(6);
    lengths[0] = 18;
    lengths[1] = 18;
    lengths[2] = 18;
    lengths[3] = 18;
    lengths[4] = 18;
    lengths[5] = 18;
    gset->setPrimLengths(lengths);
    debugVerts = (pfVec3*)pfMalloc(132*sizeof(pfVec3),pfGetSharedArena());
    for(int i=0; i<132; i++)
        {
        debugVerts[i].vec[0] = 0.0;
        debugVerts[i].vec[1] = 0.0;
        debugVerts[i].vec[2] = 0.0;
        }
    gset->setAttr(PFGS_COORD3,PFGS_PER_VERTEX,debugVerts,NULL);
    pfVec4 *colorAttr = (pfVec4*)pfMalloc(6*sizeof(pfVec4),pfGetSharedArena());
    colorAttr[0].set(1,0,0,1);
    colorAttr[1].set(1,0,0,1);
    colorAttr[2].set(0,1,0,1);
    colorAttr[3].set(0,1,0,1);
    colorAttr[4].set(0,0,1,1);
    colorAttr[5].set(0,0,1,1);
    gset->setAttr(PFGS_COLOR4,PFGS_PER_PRIM,colorAttr,NULL);
    gset->setLineWidth(0.5);
    gset->setDrawMode(PFGS_WIREFRAME,1);
    gset->setBound(NULL,PFBOUND_DYNAMIC);
    pfGeoState* gstate = new pfGeoState;
    gstate->setMode(PFSTATE_ENWIREFRAME,PF_ON);
    gset->setGState(gstate);
    geode->addGSet(gset);
    //create axis geoset
    gset = new pfGeoSet;
    lengths = (int*)pfMalloc(6*sizeof(int),pfGetSharedArena());
    gset->setPrimType(PFGS_POLYS);
    gset->setNumPrims(6);
    lengths[0] = 4;
    lengths[1] = 4;
    lengths[2] = 4;
    lengths[3] = 4;
    lengths[4] = 4;
    lengths[5] = 4;
    gset->setPrimLengths(lengths);
    gset->setAttr(PFGS_COORD3,PFGS_PER_VERTEX,debugVerts+6*18,NULL);
    gset->setAttr(PFGS_COLOR4,PFGS_OVERALL,debugColor(),NULL);
    gset->setLineWidth(1.0);
    gset->setDrawMode(PFGS_WIREFRAME,1);
    gset->setBound(NULL,PFBOUND_DYNAMIC);
    gset->setGState(gstate);
    geode->addGSet(gset);
    setDebugNode((pfGroup*)geode);
    setDebugCoords(0.5);
    }

void valuator::setDebugCoords(float radius)
    {
    // CONE_LENGTH = sqrt(CONE_RADIUS^2 + CONE_RADIUS^2)/2
    pfVec3* verts = debugVerts;
    float x_base = radius + CONE_BASE;
    float y_base = radius + CONE_BASE;
    float z_base = radius + CONE_BASE;
    float x_top = radius + CONE_TOP;
    float y_top = radius + CONE_TOP;
    float z_top = radius + CONE_TOP;
    float x_radius = CONE_RADIUS;
    float y_radius = CONE_RADIUS;
    float z_radius = CONE_RADIUS;
    float x_length = CONE_LENGTH;
    float y_length = CONE_LENGTH;
    float z_length = CONE_LENGTH;
    float x_width = AXIS_WIDTH;
    float y_width = AXIS_WIDTH;
    float z_width = AXIS_WIDTH;
    //cone vertices
    verts[0].vec[0] = x_base; verts[0].vec[1] = y_radius;
    verts[1].vec[0] = x_top;
    verts[2].vec[0] = x_base; verts[2].vec[1] = y_length; verts[2].vec[2] = -z_length;
    verts[3].vec[0] = x_top;
    verts[4].vec[0] = x_base; verts[4].vec[2] = -z_radius;
    verts[5].vec[0] = x_top;
    verts[6].vec[0] = x_base; verts[6].vec[1] = -y_length; verts[6].vec[2] = -z_length;
    verts[7].vec[0] = x_top;
    verts[8].vec[0] = x_base; verts[8].vec[1] = -y_radius;
    verts[9].vec[0] = x_top;
    verts[10].vec[0] = x_base; verts[10].vec[1] = -y_length; verts[10].vec[2] = z_length;
    verts[11].vec[0] = x_top;
    verts[12].vec[0] = x_base; verts[12].vec[2] = z_radius;
    verts[13].vec[0] = x_top;
    verts[14].vec[0] = x_base; verts[14].vec[1] = y_length; verts[14].vec[2] = z_length;
    verts[15].vec[0] = x_top;
    verts[16].vec[0] = x_base; verts[16].vec[1] = y_radius;
    verts[17].vec[0] = x_top;
    verts[18].vec[0] = -x_base; verts[18].vec[1] = y_radius;
    verts[19].vec[0] = -x_top;
    verts[20].vec[0] = -x_base; verts[20].vec[1] = y_length; verts[20].vec[2] = z_length;
    verts[21].vec[0] = -x_top;
    verts[22].vec[0] = -x_base; verts[22].vec[2] = z_radius;
    verts[23].vec[0] = -x_top;
    verts[24].vec[0] = -x_base; verts[24].vec[1] = -y_length; verts[24].vec[2] = z_length;
    verts[25].vec[0] = -x_top;
    verts[26].vec[0] = -x_base; verts[26].vec[1] = -y_radius;
    verts[27].vec[0] = -x_top;
    verts[28].vec[0] = -x_base; verts[28].vec[1] = -y_length; verts[28].vec[2] = -z_length;
    verts[29].vec[0] = -x_top;
    verts[30].vec[0] = -x_base; verts[30].vec[2] = -z_radius;
    verts[31].vec[0] = -x_top;
    verts[32].vec[0] = -x_base; verts[32].vec[1] = y_length; verts[32].vec[2] = -z_length;
    verts[33].vec[0] = -x_top;
    verts[34].vec[0] = -x_base; verts[34].vec[1] = y_radius;
    verts[35].vec[0] = -x_top;
    verts[36].vec[1] = y_base; verts[36].vec[2] = -z_radius;
    verts[37].vec[1] = y_top;
    verts[38].vec[0] = x_length; verts[38].vec[1] = y_base; verts[38].vec[2] = -z_length;
    verts[39].vec[1] = y_top;
    verts[40].vec[0] = x_radius; verts[40].vec[1] = y_base;
    verts[41].vec[1] = y_top;
    verts[42].vec[0] = x_length; verts[42].vec[1] = y_base; verts[42].vec[2] = z_length;
    verts[43].vec[1] = y_top;
    verts[44].vec[1] = y_base; verts[44].vec[2] = z_radius;
    verts[45].vec[1] = y_top;
    verts[46].vec[0] = -x_length; verts[46].vec[1] = y_base; verts[46].vec[2] = z_length;
    verts[47].vec[1] = y_top;
    verts[48].vec[0] = -x_radius; verts[48].vec[1] = y_base;
    verts[49].vec[1] = y_top;
    verts[50].vec[0] = -x_length; verts[50].vec[1] = y_base; verts[50].vec[2] = -z_length;
    verts[51].vec[1] = y_top;
    verts[52].vec[1] = y_base; verts[52].vec[2] = -z_radius;
    verts[53].vec[1] = y_top;
    verts[54].vec[1] = -y_base; verts[54].vec[2] = z_radius;
    verts[55].vec[1] = -y_top;
    verts[56].vec[0] = x_length; verts[56].vec[1] = -y_base; verts[56].vec[2] = z_length;
    verts[57].vec[1] = -y_top;
    verts[58].vec[0] = x_radius; verts[58].vec[1] = -y_base;
    verts[59].vec[1] = -y_top;
    verts[60].vec[0] = x_length; verts[60].vec[1] = -y_base; verts[60].vec[2] = -z_length;
    verts[61].vec[1] = -y_top;
    verts[62].vec[1] = -y_base; verts[62].vec[2] = -z_radius;
    verts[63].vec[1] = -y_top;
    verts[64].vec[0] = -x_length; verts[64].vec[1] = -y_base; verts[64].vec[2] = -z_length;
    verts[65].vec[1] = -y_top;
    verts[66].vec[0] = -x_radius; verts[66].vec[1] = -y_base;
    verts[67].vec[1] = -y_top;
    verts[68].vec[0] = -x_length; verts[68].vec[1] = -y_base; verts[68].vec[2] = z_length;
    verts[69].vec[1] = -y_top;
    verts[70].vec[1] = -y_base; verts[70].vec[2] = z_radius;
    verts[71].vec[1] = -y_top;
    verts[72].vec[1] = y_radius; verts[72].vec[2] = z_base;
    verts[73].vec[2] = z_top;
    verts[74].vec[0] = x_length; verts[74].vec[1] = y_length; verts[74].vec[2] = z_base;
    verts[75].vec[2] = z_top;
    verts[76].vec[0] = x_radius; verts[76].vec[2] = z_base;
    verts[77].vec[2] = z_top;
    verts[78].vec[0] = x_length; verts[78].vec[1] = -y_length; verts[78].vec[2] = z_base;
    verts[79].vec[2] = z_top;
    verts[80].vec[1] = -y_radius; verts[80].vec[2] = z_base;
    verts[81].vec[2] = z_top;
    verts[82].vec[0] = -x_length; verts[82].vec[1] = -y_length; verts[82].vec[2] = z_base;
    verts[83].vec[2] = z_top;
    verts[84].vec[0] = -x_radius; verts[84].vec[2] = z_base;
    verts[85].vec[2] = z_top;
    verts[86].vec[0] = -x_length; verts[86].vec[1] = y_length; verts[86].vec[2] = z_base;
    verts[87].vec[2] = z_top;
    verts[88].vec[1] = y_radius; verts[88].vec[2] = z_base;
    verts[89].vec[2] = z_top;
    verts[90].vec[1] = -y_radius; verts[90].vec[2] = -z_base;
    verts[91].vec[2] = -z_top;
    verts[92].vec[0] = x_length; verts[92].vec[1] = -y_length; verts[92].vec[2] = -z_base;
    verts[93].vec[2] = -z_top;
    verts[94].vec[0] = x_radius; verts[94].vec[2] = -z_base;
    verts[95].vec[2] = -z_top;
    verts[96].vec[0] = x_length; verts[96].vec[1] = y_length; verts[96].vec[2] = -z_base;
    verts[97].vec[2] = -z_top;
    verts[98].vec[1] = y_radius; verts[98].vec[2] = -z_base;
    verts[99].vec[2] = -z_top;
    verts[100].vec[0] = -x_length; verts[100].vec[1] = y_length; verts[100].vec[2] = -z_base;
    verts[101].vec[2] = -z_top;
    verts[102].vec[0] = -x_radius; verts[102].vec[2] = -z_base;
    verts[103].vec[2] = -z_top;
    verts[104].vec[0] = -x_length; verts[104].vec[1] = -y_length; verts[104].vec[2] = -z_base;
    verts[105].vec[2] = -z_top;
    verts[106].vec[1] = -y_radius; verts[106].vec[2] = -z_base;
    verts[107].vec[2] = -z_top;
    //axis vertices
    verts[108].vec[0] = -x_base; verts[108].vec[1] = -y_width;
    verts[109].vec[0] = x_base; verts[109].vec[1] = -y_width;
    verts[110].vec[0] = x_base; verts[110].vec[1] = y_width;
    verts[111].vec[0] = -x_base; verts[111].vec[1] = y_width;
    verts[112].vec[0] = -x_base; verts[112].vec[2] = -z_width;
    verts[113].vec[0] = x_base; verts[113].vec[2] = -z_width;
    verts[114].vec[0] = x_base; verts[114].vec[2] = z_width;
    verts[115].vec[0] = -x_base; verts[115].vec[2] = z_width;
    verts[116].vec[1] = -y_base; verts[116].vec[2] = -z_width;
    verts[117].vec[1] = y_base; verts[117].vec[2] = -z_width;
    verts[118].vec[1] = y_base; verts[118].vec[2] = z_width;
    verts[119].vec[1] = -y_base; verts[119].vec[2] = z_width;
    verts[120].vec[0] = -x_width; verts[120].vec[1] = -y_base;
    verts[121].vec[0] = -x_width; verts[121].vec[1] = y_base;
    verts[122].vec[0] = x_width; verts[122].vec[1] = y_base;
    verts[123].vec[0] = x_width; verts[123].vec[1] = -y_base;
    verts[124].vec[0] = -x_width; verts[124].vec[2] = -z_base;
    verts[125].vec[0] = -x_width; verts[125].vec[2] = z_base;
    verts[126].vec[0] = x_width; verts[126].vec[2] = z_base;
    verts[127].vec[0] = x_width; verts[127].vec[2] = -z_base;
    verts[128].vec[1] = -y_width; verts[128].vec[2] = -z_base;
    verts[129].vec[1] = -y_width; verts[129].vec[2] = z_base;
    verts[130].vec[1] = y_width; verts[130].vec[2] = z_base;
    verts[131].vec[1] = y_width; verts[131].vec[2] = -z_base;
    pfBox box;
    box.min.set(-x_top,-y_top,-z_top);
    box.max.set(x_top,y_top,z_top);
    ((pfGeode*)debugNode())->getGSet(0)->setBound(&box,PFBOUND_STATIC);
    ((pfGeode*)debugNode())->getGSet(1)->setBound(&box,PFBOUND_STATIC);
    debugNode()->setTravMask(PFTRAV_ISECT, YG_ISECT_ALLMASK|YG_ISECT_DEBUGMASK, PFTRAV_SELF|PFTRAV_DESCEND|PFTRAV_IS_CACHE, PF_SET);
    pfSphere sphere;
    sphere.radius = radius + CONE_TOP;
    sphere.center.set(0,0,0);
    debugNode()->setBound(NULL,PFBOUND_DYNAMIC);
    }
