// Description: a node that sets the visual priority of children
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> lower depth values result in higher visual priority
//<li> you must add the following to your .ygrc file in order to use this node:<br>
//<ul>
//<li> setenv YG_PRELOAD_CLASSES visualPriority
//</ul>
//</ul>
//
// Category: Geometry
// Author: Alex Hill
// Revision: 11/10/01
//
#include <Performer/pf/pfTraverser.h>
#include <ygNodeDB.h>
#include <ygNetMessages.h>
#include "visualPriority.h"


extern "C" ygNode* APIENTRY construct_visualPriority(const char* name,bool master) { return new visualPriority(name,master); }

typedef struct _priorityData
    {
    bool active;
    int depth;
    } priorityData_t;

visualPriority::visualPriority(const char* name,bool master) : ygNode(name,master)
    {
    setClassName("visualPriority",true);
    //allocate shared memory data structure for pre and post callbacks
    data_ = (priorityData_t*) pfMalloc(sizeof(priorityData_t),pfGetSharedArena());
    pfnode()->setTravFuncs(PFTRAV_DRAW,preDrawFunc,postDrawFunc);
    pfnode()->setTravData(PFTRAV_DRAW,data_);
    data_->active = true;
    data_->depth = 0;
    //distribute the off state of the visualPriority buffer
    addNetMessage("off",&data_->active,YG_NET_BOOL_NEGATE);
    //distribute the visualPriority depth
    addNetMessage("depth",&data_->depth,YG_NET_INT);
    }

visualPriority::~visualPriority(void)
    {
    }

void visualPriority::reset(void)
    {
    //clear nodeNames string
    resetNetMessage("depth");
    ygNode::reset();
    }

void visualPriority::message(const ygMessage& msg)
    {
    //set the visual priority
    if (msg == "depth")
        {
        return;
        }
    //turn the priority on
    else if (msg == "on")
        {
        data_->active = true;
        netMessageChanged("off");
        }
    //turn the priority off
    else if (msg == "off")
        {
        return;
        }
    else
        ygNode::message(msg);
    }

int visualPriority::preDrawFunc(pfTraverser* trav,void* data)
    {
    //cast the traversal data as the priorityData structure
    priorityData_t* priorityData = (priorityData_t*)data;
    //if the plane is active then
    if (priorityData->active)
        {
        float low = priorityData->depth/(float)65636;
        float high = low + 1/(float)65636;
        glDepthRange(low,high);
        }
    return PFTRAV_CONT;
    }
    
int visualPriority::postDrawFunc(pfTraverser* trav,void* data)
    {
    //cast the traversal data as the priorityData structure
    priorityData_t* priorityData = (priorityData_t*)data;
    //if the plane is active then
    if (priorityData->active)
        glDepthRange(0.0,1.0);
    return PFTRAV_CONT;
    }


