// Description: creates a positional whiteNoise source within a space
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> the volume of a sound is the region in which the sound will play at 100% amplitude(loudness)
//<li> falloffDistance is a region outside the volume in which the amplitude is diminished as the user moves further away from the source(volume)
//<li> the sound server, snerd, must be run on the a local or remote machine
//<li> sounds under a switch will turn off but must be restarted when the switch is toggled back on
//</ul>
//
// Category: Sound
// Author: Alex Hill
// Revision: 08/01/03
//
#include <pfcave.h>
#include <bergenWhiteNoise.h>
#include "ygWorld.h"
#include "ygNetMessages.h"
#include "whiteNoise.h"

using namespace std;

extern "C" ygNode* APIENTRY construct_whiteNoise(const char* name,bool master) { return new whiteNoise(name,master); }

struct _whiteNoisePrivateData
    {
    bergenWhiteNoise* sample;
    bool isPlaying;
    bool netPlaying;
    bool isPaused;
    float curAmplitude;
    float distance;
    pfVec3 currPosition;
    float updateInterval, lastUpdate;
    };


whiteNoise::whiteNoise(const char* name,bool master) : ygSoundServer(name,master)
    {
    setClassName("whiteNoise",true);
    p_ = new struct _whiteNoisePrivateData;
    p_->sample = NULL;
    p_->curAmplitude = 0;
    p_->updateInterval = 0.1;
    if (getenv("YG_SOUND_INTERVAL"))
        p_->updateInterval = atof(getenv("YG_SOUND_INTERVAL"));
    p_->lastUpdate = 0;
    p_->isPlaying = false;
    p_->isPaused = false;
    p_->netPlaying = false;
    addNetMessage("play", &p_->netPlaying, YG_NET_BOOL_NEGATE);
    addNetMessage("stop", &p_->netPlaying, YG_NET_BOOL);
    dontSaveNetMessage("stop");
    addNetMessage("updateInterval",&p_->updateInterval,YG_NET_FLOAT);
    }

whiteNoise::~whiteNoise(void)
    {
    if (p_->sample)
        delete p_->sample;
    delete p_;
    }


void whiteNoise::reset(void)
    {
    //reset playing flag to false
    resetNetMessage("play");
    p_->isPlaying = false;
    p_->isPaused = false;
    //set current amplitude to 0.0
    p_->curAmplitude = 0;
    //set update interval to 0.1
    resetNetMessage("updateInterval");
    //if a sample exists then delete it
    if (p_->sample)
        {
        stop();
        delete p_->sample;
        p_->sample = NULL;
        }
    ygSpace::reset();
    }


void whiteNoise::message(const ygMessage& msg)
    {
    //start the sound playing
    if (msg == "play")
        {
        {
        if (p_->netPlaying)
            play();
        else
            stop();
        }
        }
    //stop the playing sound
    else if (msg == "stop")
        {
        if (p_->netPlaying)
            play();
        else
            stop();
        }
    //set the update interval for sound server
    else if (msg == "updateInterval")
        {
        return;
        }
    else
        ygSoundServer::message(msg);
    }


void whiteNoise::setDistance(float val)
    {
    if (p_->sample)
        p_->sample->setDistance(val);
    p_->distance = val;
    }


void whiteNoise::setAmplitude(float val)
    {
    if (p_->sample)
        p_->sample->setAmplitude(val);
    ygSoundServer::setAmplitude(val);
    }


void whiteNoise::setFalloffDistance(float val)
    {
    if (p_->sample)
        p_->sample->setFalloffDistance(val);
    ygSoundServer::setFalloffDistance(val);
    }


void whiteNoise::setAttenuation(int val)
    {
    if (p_->sample)
        p_->sample->setAttenuationType((AttenuationType)val);
    ygSoundServer::setAttenuation(val);
    }


bool whiteNoise::isPlaying(void)
    {
    return p_->isPlaying;
    }


bool whiteNoise::isPaused(void)
    {
    return p_->isPaused;
    }


void whiteNoise::play(void)
    {
    if (p_->isPlaying)
        return;
    //if no sample then create a new bergenSample
    if (!p_->sample)
        p_->sample = new bergenWhiteNoise(server());
    p_->sample->setAmplitude(amplitude());
    p_->sample->setFalloffDistance(falloffDistance());
    p_->sample->setAttenuationType((AttenuationType)attenuation());
    setDistance(computeDistance());
    //play the sample
    p_->sample->play();
    //generate a heartbeat for the server
    heartbeat();
    p_->isPlaying = true;
    eventOccurred("start");
    }


void whiteNoise::pause(void)
    {
    if (p_->isPlaying)
        {
        p_->isPaused = true;
        p_->isPlaying = false;
        p_->sample->pause();
        }
    }


void whiteNoise::stop(bool sendStopCommand)
    {
    //if a sample is playing
    if (p_->isPlaying || p_->isPaused)
        {
        //send a stop command to the server
        if ((sendStopCommand) && (p_->sample))
            p_->sample->stop();
        p_->isPlaying = false;
        p_->isPaused = false;
        p_->netPlaying = false;
        eventOccurred("stop");
        }
    }


void whiteNoise::app(void)
    {
    //if a sample exists
    if (p_->sample)
        {
        //generate a heartbeat for the server
        heartbeat();
        //if the sample is playing
        if (p_->isPlaying)
            {
            //if update interval has passed then update sound server
            if ((ygWorld::frameTime() - p_->lastUpdate) > p_->updateInterval)
                {
                float newDistance = computeDistance();
                if (newDistance != p_->distance)
                    setDistance(newDistance);
                //determine position of this sound
                pfVec3 newPos;
                newPos = origin();
                pfVec3 cavePos;
                //get position relative to CAVE
                CAVENavConvertWorldToCAVE(newPos.vec,cavePos.vec);
                if (cavePos != p_->currPosition)
                    {
                    if (p_->sample)
                        p_->sample->setPosition(cavePos.vec[0],cavePos.vec[2],-cavePos.vec[1]);
                    p_->currPosition = cavePos;
                    }
                p_->lastUpdate = ygWorld::frameTime();
                }
            }
        }
    ygSoundServer::app();
    }
