// Description: establishes sky color, fog, and clipping plane parameters within a space 
// 
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> this node can also be created with the alias: environment
//<li> volumes are defined by a box, cylinder, sphere, point or infinite volume 
//<li> for volume messages refer to the <a href="ygSpace.html">ygSpace</a> node
//<li> environment settings are applied when the user enters the space                        
//<li> the last environment defined will be applied when environments overlap
//<li> objects located outside of the field defined by the clipping planes will not be drawn
//<li> fog is the atmospheric effect of aerosol water particles that occludes vision over a distance
//<li> fog color is blended with the pixel color based on the range from the viewport and the fog function
//<li> the fog function is specified by the following "type" options:
//<ul>
//<li> linear (computes fog linearly at pixels),
//<li> exp (computes fog exponentially at pixels (ex)),
//<li> exp2 (computes fog exponentially at pixels (ex squared))
//</ul>
//<li> "onset" is the distance from the viewport at which the fog begins to take effect.
//<li> "opaque" is the distance at which the fog is 100% opaque, objects beyond this distance will not be visible
//</ul>
// 
// Category: Attributes
// Author: Dave Pape
//           11/01/01
//           06/01/03 Alex Hill - added acceptNetMessage function to update clients
// Revision: 08/01/03 Dave Pape - updated logic to fix overlapping volumes
//
#include "ygUtil.h"
#include "ygWorld.h"
#include "ygNetMessages.h"
#include "ygEnvironment.h"

#include "ygDLL.h"

using namespace std;

extern "C" YG_DLL_API ygNode* APIENTRY construct_ygEnvironment(const char* name,bool net) { return new ygEnvironment(name,net); }

struct _ygEnvironmentPrivateData
    {
    pfVec2 clip;
    pfVec3 skyColor;
    bool fog;
    int fogType;
    pfVec3 fogColor;
    float fogOnset, fogOpaque;
    int lastApply;
    };


ygEnvironment * ygEnvironment::activeEnvironment_ = NULL;


ygEnvironment::ygEnvironment(const char* name,bool net) : ygSpace(name,net)
    {
    setClassName("ygEnvironment",true);
    p_ = new struct _ygEnvironmentPrivateData;
    p_->clip[0] = 0.1;
    p_->clip[1] = 1000;
    p_->skyColor.set(0,0,0);
    p_->fogType = PFFOG_OFF;
    p_->fogColor.set(1,1,1);
    p_->fogOnset = 0.0;
    p_->fogOpaque = 100.0;
    p_->lastApply = -999;
    addNetMessage("clip",&p_->clip,YG_NET_VEC2);
    addNetMessage("skyColor",&p_->skyColor,YG_NET_VEC3);
    addNetMessage("fog",&p_->fogType,YG_NET_INT);
    addNetMessageMap("fog","off",PFFOG_OFF);
    addNetMessageMap("fog","lin",PFFOG_PIX_LIN);
    addNetMessageMap("fog","linear",PFFOG_PIX_LIN);
    addNetMessageMap("fog","exp",PFFOG_PIX_EXP);
    addNetMessageMap("fog","exp2",PFFOG_PIX_EXP2);
    addNetMessage("fog", &p_->fogColor, YG_NET_VEC3);
    addNetMessage("fog", &p_->fogOnset, YG_NET_FLOAT);
    addNetMessage("fog", &p_->fogOpaque, YG_NET_FLOAT);
    }

ygEnvironment::~ygEnvironment(void)
    {
    p_->clip[0] = 0.1;
    p_->clip[1] = 1000;
    p_->skyColor.set(0,0,0);
    p_->fogType = PFFOG_OFF;
    p_->fogColor.set(1,1,1);
    p_->fogOnset = 0.0;
    p_->fogOpaque = 100.0;
    p_->lastApply = -999;
    apply();
    if (activeEnvironment_ = this)
        activeEnvironment_ = NULL;
    delete p_;
    }


void ygEnvironment::reset(void)
    {
    //set near and far planes to 0.1 and 1000
    resetNetMessage("clip");
    setClip(p_->clip[0],p_->clip[1]);
    //set sky color to black (0,0,0)
    resetNetMessage("skycolor");
    setSkyColor(p_->skyColor);
    //set fog to off
    resetNetMessage("fog");
    setFog(p_->fogType,p_->fogColor,p_->fogOnset,p_->fogOpaque);
    ygSpace::reset();
    }


void ygEnvironment::message(const ygMessage& msg)
    {
    //set the near and far clipping planes
    if (msg == "clip")
        {
        setClip(p_->clip[0],p_->clip[1]);
        }
    //set the sky color
    else if (msg == "skyColor")
        {
        setSkyColor(p_->skyColor);
        }
    //set the fog properties (type, color, onset, opaque)
    else if (msg == "fog")
        {
        setFog(p_->fogType,p_->fogColor,p_->fogOnset,p_->fogOpaque);
        }
    else
        ygSpace::message(msg);
    }


#ifdef WIN32
// In visual C++, near and far are reserved words for specifying near and far pointers
void ygEnvironment::setClip(float n,float f)
    {
    p_->clip[0] = n;
    p_->clip[1] = f;
    p_->lastApply = -999;
    }
#else
void ygEnvironment::setClip(float near,float far)
    {
    p_->clip[0] = near;
    p_->clip[1] = far;
    p_->lastApply = -999;
    }
#endif

float ygEnvironment::nearClip(void) const
    {
    return p_->clip[0];
    }


float ygEnvironment::farClip(void) const
    {
    return p_->clip[1];
    }


void ygEnvironment::setSkyColor(const pfVec3& color)
    {
    p_->skyColor = color;
    p_->lastApply = -999;
    }


const pfVec3& ygEnvironment::skyColor(void) const
    {
    return p_->skyColor;
    }


void ygEnvironment::fogOff(void)
    {
    p_->fog = false;
    p_->lastApply = -999;
    }


void ygEnvironment::setFog(int fogtype,const pfVec3& color,float onset,float opaque)
    {
    p_->fog = true;
    p_->fogType = fogtype;
    p_->fogColor = color;
    p_->fogOnset = onset;
    p_->fogOpaque = opaque;
    p_->lastApply = -999;
    }


void ygEnvironment::app(void)
    {
    ygNode * viewer = ygWorld::world()->viewer();
    //if space contains user then apply settings
    if ((viewer) && (contains(viewer->origin(this))))
        apply();
    ygSpace::app();
    }


void ygEnvironment::apply(void)
    {
    //if properties have changed
    if (p_->lastApply < ygWorld::frameNumber()-1)
        {
        //set world clipping planes
        ygWorld::world()->setClip(p_->clip[0], p_->clip[1]);
        //set world sky color
        ygWorld::world()->setSkyColor(p_->skyColor);
        //if fog then set world fog properties
        if (p_->fogType)
            ygWorld::world()->setFog(p_->fogType, p_->fogColor,
                        p_->fogOnset, p_->fogOpaque);
        else
            ygWorld::world()->fogOff();
        }
    /* The activeEnvironment_ class variable is intended to take care
      of cases where environment nodes overlap.  It keeps track of which
      node is 'active', i.e. the last one to apply itself.  Effectively,
      when there is an overlap, the winning node changes the other node's
      lastApply value to let the other node know that it has been un-applied,
      allowing it to re-apply itself when the user gets out of the overlap
      region. */
    if (activeEnvironment_)
        activeEnvironment_->p_->lastApply = -999;
    activeEnvironment_ = this;
    p_->lastApply = ygWorld::frameNumber();
    }
