// Description: creates a positional sound within a space using one of several file formats
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> this node can also be created with the alias: sound
//<li> the volume of a sound is the region in which the sound will play at 100% amplitude(loudness)
//<li> falloffDistance is a region outside the volume in which the amplitude is diminished as the user moves further away from the source(volume)
//<li> Common file formats supported by snerd:
//<b> .aiff</b> - Audio Interchange File Format
//<b> .wav</b> - MS RIFF WAVE Format
//<li> sound files should be saved using the <b>.aiff</b> format for best performance
//<li> the sound server, snerd, must be run on the a local or remote machine
//<li> the location of sound files is set by the YG_PATH environmental variable
//<li> sounds under a switch will turn off but must be restarted when the switch is toggled back on
//</ul>
//
// Category: Sound
// Author: Dave Pape
//           11/01/01
//           10/03/02 Dave Pape - Added 'start' event
// Revision: 08/01/03 Alex Hill - added position reporting for spatialized sound server
//
#include <pfcave.h>
#include <bergenSample.h>
#include "ygWorld.h"
#include "ygNetMessages.h"
#include "ygSound.h"

#include "ygDLL.h"

using namespace std;

extern "C" YG_DLL_API ygNode* APIENTRY construct_ygSound(const char* name,bool master) { return new ygSound(name,master); }

struct _ygSoundPrivateData
    {
    ygString filename;
    bergenSample* sample;
    bool isPlaying;
    bool netPlaying;
    bool isPaused;
    bool loop;
    float distance;
    pfVec3 currPosition;
    float updateInterval, lastUpdate;
    };


ygSound::ygSound(const char* name,bool master) : ygSoundServer(name,master)
    {
    setClassName("ygSound",true);
    p_ = new struct _ygSoundPrivateData;
    p_->sample = NULL;
    p_->loop = false;
    p_->distance = 0;
    p_->updateInterval = 0.1;
    if (getenv("YG_SOUND_INTERVAL"))
        p_->updateInterval = atof(getenv("YG_SOUND_INTERVAL"));
    p_->lastUpdate = 0;
    p_->isPlaying = false;
    p_->isPaused = false;
    p_->netPlaying = false;
    addNetMessage("play", &p_->netPlaying,YG_NET_BOOL_NEGATE);
    addNetMessage("stop", &p_->netPlaying,YG_NET_BOOL);
    dontSaveNetMessage("stop");
    addNetMessage("file", &p_->filename,YG_NET_FILE);
    addNetMessage("loop", &p_->loop,YG_NET_BOOL_NEGATE);
    addNetMessage("updateInterval",&p_->updateInterval,YG_NET_FLOAT);
    }

ygSound::~ygSound(void)
    {
    if (p_->sample)
        delete p_->sample;
    delete p_;
    }


void ygSound::reset(void)
    {
    //reset playing flag to false
    resetNetMessage("play");
    p_->isPlaying = false;
    p_->isPaused = false;
    //reset loop flag to false
    resetNetMessage("loop");
    //set current distance to 0.0
    p_->distance = 0;
    //set update interval to 0.1
    resetNetMessage("updateInterval");
    //if a sample exists then delete it
    if (p_->sample)
        {
        stop();
        delete p_->sample;
        p_->sample = NULL;
        }
    resetNetMessage("file");
    ygSoundServer::reset();
    }


void ygSound::message(const ygMessage& msg)
    {
    //set the name of the sound file
    if (msg == "file")
        {
        setFile(p_->filename);
        }
    //set the sound looping mode
    else if (msg == "loop")
        {
        setLoop(p_->loop);
        }
    //start the sound playing with file and amplitude
    else if (msg == "play")
        {
        if (p_->netPlaying)
            play();
        else
            stop();
        }
    //stop the sound playing with file and amplitude
    else if (msg == "stop")
        {
        if (p_->netPlaying)
            play();
        else
            stop();
        }
    //set the update interval for sound server
    else if (msg == "updateInterval")
        {
        return;
        }
    else
        ygSoundServer::message(msg);
    }


void ygSound::setFile(const ygString& val)
    {
    if (p_->sample)
        {
        stop();
        delete p_->sample;
        p_->sample = NULL;
        }
    p_->filename = val;
    }


void ygSound::setLoop(bool val)
    {
    if (p_->sample)
        p_->sample->setLoop(val);
    p_->loop = val;
    }


void ygSound::setDistance(float val)
    {
    if (p_->sample)
        p_->sample->setDistance(val);
    p_->distance = val;
    }


void ygSound::setAmplitude(float val)
    {
    if (p_->sample)
        p_->sample->setAmplitude(val);
    ygSoundServer::setAmplitude(val);
    }


void ygSound::setFalloffDistance(float val)
    {
    if (p_->sample)
        p_->sample->setFalloffDistance(val);
    ygSoundServer::setFalloffDistance(val);
    }


void ygSound::setAttenuation(int val)
    {
    if (p_->sample)
        p_->sample->setAttenuationType((AttenuationType)val);
    ygSoundServer::setAttenuation(val);
    }


bool ygSound::isPlaying(void)
    {
    return p_->isPlaying;
    }


bool ygSound::isPaused(void)
    {
    return p_->isPaused;
    }


void ygSound::play(void)
    {
    if (p_->isPaused)
        {
        p_->isPlaying = true;
        p_->isPaused = false;
        p_->sample->play();
        }
    if (p_->isPlaying)
        return;
    //if no sample then create a new bergenSample
    if (!p_->sample)
        p_->sample = new bergenSample(p_->filename.c_str(),server());
    p_->sample->setLoop((int)p_->loop);
    p_->sample->setAmplitude(amplitude());
    p_->sample->setFalloffDistance(falloffDistance());
    p_->sample->setAttenuationType((AttenuationType)attenuation());
    setDistance(computeDistance());
    p_->sample->play();
    //generate a heartbeat for the server
    heartbeat();
    p_->isPlaying = true;
    //the sound has started
    eventOccurred("start");
    }


void ygSound::pause(void)
    {
    if (p_->isPlaying)
        {
        p_->isPaused = true;
        p_->isPlaying = false;
        p_->sample->pause();
        }
    }


void ygSound::stop(bool sendStopCommand)
    {
    //if a sample is playing
    if (p_->isPlaying || p_->isPaused)
        {
        //send a stop command to the server
        if ((sendStopCommand) && (p_->sample))
            p_->sample->stop();
        p_->isPlaying = false;
        p_->isPaused = false;
        p_->netPlaying = false;
        //the sound has stopped
        eventOccurred("stop");
        }
    }


void ygSound::app(void)
    {
    //if a sample exists
    if (p_->sample)
        {
        //generate a heartbeat for the server
        heartbeat();
        //if the sample is playing
        if (p_->isPlaying)
            {
            //if sample has stopped then stop
            if ((!p_->loop) && (!p_->sample->isPlaying()))
                stop(false);
            //else, compute new amplitude
            else
                {
                //if update interval has passed then update sound server
                if ((ygWorld::frameTime() - p_->lastUpdate) > p_->updateInterval)
                    {
                    float newDistance = computeDistance();
                    if (newDistance != p_->distance)
                        setDistance(newDistance);
                    //determine position of this sound
                    pfVec3 newPos;
                    newPos = origin();
                    pfVec3 cavePos;
                    //get position relative to CAVE
                    CAVENavConvertWorldToCAVE(newPos.vec,cavePos.vec);
                    if (cavePos != p_->currPosition)
                        {
                        if (p_->sample)
                            p_->sample->setPosition(cavePos.vec[0],cavePos.vec[2],-cavePos.vec[1]);
                        p_->currPosition = cavePos;
                        }
                    p_->lastUpdate = ygWorld::frameTime();
                    }
                }
            }
        }
    ygSoundServer::app();
    }
