// Description: creates a bergen sound server and manages sound clients
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> this node is the base class for most sound related nodes
//<li> the volume of a sound is the region in which the sound will play at 100% amplitude(loudness)
//<li> falloffDistance is a region outside the volume in which the amplitude is diminished as the user moves further away from the source(volume)
//<li> Common file formats supported by snerd:
//<b> .aiff</b> - Audio Interchange File Format
//<b> .wav</b> - MS RIFF WAVE Format
//<li> sound files should be saved using the <b>.aiff</b> format for best performance
//<ol> 
//<li> the sfconvert utility is located in /usr/sbin/
//<li> type sfconvert then the name of the file you want to convert, followed by the name of the converted file
//<li> <b>ex:</b> sfconvert oldfile.wav newfile.aif
//<li> see the man pages on sfconvert for more information
//</ol>
//<li> the sound server, snerd, must be run on the a local or remote machine
//<li> all sounds should be included in the YG_PATH environmental variable
//<li> sounds under a switch will turn off but must be restarted when the switch is toggled back on
//</ul>
//
// Category: Foundation
// Author: Alex Hill
// Revision: 08/01/03
// Revision: 07/01/06 Alex Hill - added logic to allow pausing of sound files
//
#include <stdio.h>
#include <pfcave.h>
#include <bergenServer.h>
#include <bergenSound.h>
#include <vector>
#include "ygVolume.h"
#include "ygUtil.h"
#include "ygWorld.h"
#include "ygNetMessages.h"
#include "ygSoundServer.h"

#include "ygDLL.h"

using namespace std;

extern "C" YG_DLL_API ygNode* APIENTRY construct_ygSoundServer(const char* name,bool master) { return new ygSoundServer(name,master); }

struct _ygSoundServerPrivateData
    {
    float falloffDistance;
    float amplitude;
    int attenuation;
    pfVec3 currPosition;
    unsigned int lastHeartbeat;
    };


struct _ygSoundServerGlobalData
    {
    bergenServer* server;
    vector<ygSoundServer*> sounds;
    unsigned int currentHeartbeat;
    float updateInterval, lastUpdate;
    };

struct _ygSoundServerGlobalData* ygSoundServer::global_ = NULL;



ygSoundServer::ygSoundServer(const char* name,bool master) : ygSpace(name,master)
    {
    setClassName("ygSoundServer",true);
    if (!global_)
        initServer();
    p_ = new struct _ygSoundServerPrivateData;
    p_->falloffDistance = 10;
    p_->amplitude = 1;
    p_->attenuation = LINEAR_FALLOFF;
    addSound(this);
    addNetMessage("amplitude",&p_->amplitude,YG_NET_FLOAT);
    addNetMessage("falloffDistance",&p_->falloffDistance,YG_NET_FLOAT);
    addNetMessage("attenuation",&p_->attenuation,YG_NET_INT);
    addNetMessageMap("attenuation","none",NONE);
    addNetMessageMap("attenuation","linear",LINEAR_FALLOFF);
    addNetMessageMap("attenuation","squared",INVERSE_SQUARE_LAW);
    addEvent("start");
    addEvent("stop");
    }

ygSoundServer::~ygSoundServer(void)
    {
    removeSound(this);
    delete p_;
    }


void ygSoundServer::reset(void)
    {
    //set amplitude to 1.0
    resetNetMessage("amplitude");
    //set fall off distance to 10.0
    resetNetMessage("falloffDistance");
    //set attenuation type
    resetNetMessage("attenuation");
    ygSpace::reset();
    }


void ygSoundServer::message(const ygMessage& msg)
    {
    //set the amplitude within the space
    if (msg == "amplitude")
        {
        setAmplitude(p_->amplitude);
        }
    //set the zero amplitude distance from the space perimeter
    else if (msg == "falloffDistance")
        {
        setFalloffDistance(p_->falloffDistance);
        }
    //set distance attenuation type
    else if (msg == "attenuation")
        {
        setAttenuation(p_->attenuation);
        }
    //send a message to the server
    else if (msg == "send")
        {
        if (msg.numArgs() > 0)
            {
            char buffer[1024];
            osc::OutboundPacketStream p( buffer, 1024 );
            ygString arg;
            arg = msg.stringArg(0);
            p << osc::BeginMessage( arg.c_str() );
            for (int i=1; i < msg.numArgs(); i++)
                {
                arg = msg.stringArg(i);
                if (!isalpha(arg.c_str()[0]))
                    {
                    if (strchr(arg.c_str(),'.'))
                        {
                        float val = atof(arg.c_str());
                        p << val;
                        }
                    else
                        {
                        int val = atoi(arg.c_str());
                        p << val;
                        }
                    }
                else
                    p << arg.c_str();
                }
            p << osc::EndMessage;
            global_->server->sendMessage( p );
            }
        }
    else
        ygSpace::message(msg);
    }


void ygSoundServer::setFalloffDistance(float val)
    {
    p_->falloffDistance = val;
    }


void ygSoundServer::setAmplitude(float val)
    {
    p_->amplitude = val;
    }


void ygSoundServer::setAttenuation(int val)
    {
    p_->attenuation = val;
    }


float ygSoundServer::amplitude(void)
    {
    return p_->amplitude;
    }


float ygSoundServer::falloffDistance(void)
    {
    return p_->falloffDistance;
    }


int ygSoundServer::attenuation(void)
    {
    return p_->attenuation;
    }


bool ygSoundServer::isPlaying(void)
    {
    return 0;
    }


bool ygSoundServer::isPaused(void)
    {
    return 0;
    }


void ygSoundServer::play(void)
    {
    return;
    }


void ygSoundServer::pause(void)
    {
    return;
    }


void ygSoundServer::stop(bool sendStopCommand)
    {
    return;
    }


void ygSoundServer::heartbeat(void)
    {
    if (global_)
        p_->lastHeartbeat = global_->currentHeartbeat;
    if(p_->lastHeartbeat == 0)
        p_->lastHeartbeat = 1;
    }


void ygSoundServer::app(void)
    {
    ygSpace::app();
    }


float ygSoundServer::computeDistance(void)
    {
    float dist = 0.0;
    ygNode * viewer = ygWorld::world()->viewer();
    if (viewer)
        {
        viewer->app();
        pfVec3 userPos = viewer->origin(this);
        if (volume())
            dist = volume()->distance(userPos);
        else
            dist = sqrtf(userPos.dot(userPos));
        }
    return dist;
    }


/****** Global functions for connection to server & heartbeat *******/

void ygSoundServer::initServer(void)
    {
    if (!global_)
        {
        global_ = new struct _ygSoundServerGlobalData;
        global_->server = new bergenServer;
        if (!getenv("BERGEN_PATH"))
            {
            if (getenv("YG_PATH"))
                global_->server->addPath(getenv("YG_PATH"));
            }
        global_->currentHeartbeat = 0;
        global_->updateInterval = 0.1;
        if (getenv("YG_LISTENER_INTERVAL"))
            global_->updateInterval = atof(getenv("YG_LISTENER_INTERVAL"));
        global_->lastUpdate = 0;
        }
    }


void ygSoundServer::addSound(ygSoundServer* s)
    {
    if (global_)
        global_->sounds.push_back(s);
    }


void ygSoundServer::removeSound(ygSoundServer* s)
    {
    if (global_)
        {
        vector<ygSoundServer*>::iterator iter;
        for (iter = global_->sounds.begin(); iter != global_->sounds.end(); ++iter)
            if (*iter == s)
                {
                global_->sounds.erase(iter);
                break;
                }
        }
    }


void ygSoundServer::updateServer(void)
    {
    vector<ygSoundServer*>::iterator iter;
    for (iter = global_->sounds.begin(); iter != global_->sounds.end(); ++iter)
        {
        if (((*iter)->isPlaying()) &&
            ((*iter)->p_->lastHeartbeat < global_->currentHeartbeat))
            (*iter)->pause();
        else if (((*iter)->isPaused()) &&
            ((*iter)->p_->lastHeartbeat == global_->currentHeartbeat))
            (*iter)->play();
        }
    global_->currentHeartbeat++;
    global_->server->processMessages();
    //if update interval has passed then
    if ((ygWorld::frameTime() - global_->lastUpdate) > global_->updateInterval)
        {
        //get position relative to CAVE
        float* pos = new float[3];
        CAVEGetPosition(CAVE_HEAD,pos);
        //update sound server
        global_->server->listener()->setPosition(pos[0],pos[2],-pos[1]);
        global_->lastUpdate = ygWorld::frameTime();
        }
    }


void ygSoundServer::closeServer(void)
    {
    delete global_->server;
    delete global_;
    global_ = NULL;
    }


bergenServer * ygSoundServer::server(void)
    {
    return global_->server;
    }
